# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import (
    Menu,
    Panel,
    UIList,
)
from rna_prop_ui import PropertyPanel
from bpy.app.translations import (
    contexts as i18n_contexts,
    pgettext_iface as iface_,
    pgettext_rpt as rpt_,
)
from bl_ui.utils import PresetPanel
from bl_ui.space_properties import PropertiesAnimationMixin

from bl_ui.properties_physics_common import (
    point_cache_ui,
    effector_weights_ui,
    basic_force_field_settings_ui,
    basic_force_field_falloff_ui,
)


def particle_panel_enabled(context, psys):
    if psys is None:
        return True
    phystype = psys.settings.physics_type
    if psys.settings.type in {'EMITTER', 'REACTOR'} and phystype in {'NO', 'KEYED'}:
        return True
    else:
        return (psys.point_cache.is_baked is False) and (not psys.is_edited) and (not context.particle_system_editable)


def particle_panel_poll(cls, context):
    psys = context.particle_system
    engine = context.engine
    settings = 0

    if psys:
        settings = psys.settings
    elif isinstance(context.space_data.pin_id, bpy.types.ParticleSettings):
        settings = context.space_data.pin_id

    if not settings:
        return False

    return (settings.is_fluid is False) and (engine in cls.COMPAT_ENGINES)


def particle_get_settings(context):
    if context.particle_system:
        return context.particle_system.settings
    elif isinstance(context.space_data.pin_id, bpy.types.ParticleSettings):
        return context.space_data.pin_id
    return None


class PARTICLE_MT_context_menu(Menu):
    bl_label = "Particle Specials"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        psys = context.particle_system

        props = layout.operator(
            "particle.copy_particle_systems",
            text="Copy Active to Selected Objects",
            icon='COPYDOWN',
        )
        props.use_active = True
        props.remove_target_particles = False

        props = layout.operator(
            "particle.copy_particle_systems",
            text="Copy All to Selected Objects",
        )
        props.use_active = False
        props.remove_target_particles = True

        if psys is not None and psys.settings.type == 'HAIR':
            layout.operator("curves.convert_from_particle_system", text="Convert to Curves")

        layout.separator()

        layout.operator(
            "particle.duplicate_particle_system",
            icon='DUPLICATE',
        )
        layout.operator("particle.particle_system_remove_all", text="Remove All Particle Systems")


class PARTICLE_PT_hair_dynamics_presets(PresetPanel, Panel):
    bl_label = "Hair Dynamics Presets"
    preset_subdir = "hair_dynamics"
    preset_operator = "script.execute_preset"
    preset_add_operator = "particle.hair_dynamics_preset_add"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }


class ParticleButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "particle"

    @classmethod
    def poll(cls, context):
        return particle_panel_poll(cls, context)


def find_modifier(ob, psys):
    for md in ob.modifiers:
        if md.type == 'PARTICLE_SYSTEM':
            if md.particle_system == psys:
                return md
    return None


class PARTICLE_UL_particle_systems(UIList):

    def draw_item(self, _context, layout, data, item, icon, _active_data, _active_propname, _index, _flt_flag):
        ob = data
        psys = item

        md = find_modifier(ob, psys)
        row = layout.row(align=True)

        row.prop(psys, "name", text="", emboss=False, icon_value=icon)
        if md:
            row.prop(
                md,
                "show_viewport",
                emboss=False,
                icon_only=True,
            )
            row.prop(
                md,
                "show_render",
                emboss=False,
                icon_only=True,
            )


class PARTICLE_PT_context_particles(ParticleButtonsPanel, Panel):
    bl_label = ""
    bl_options = {'HIDE_HEADER'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        engine = context.engine
        return (
            (context.particle_system or context.object or context.space_data.pin_id) and
            (engine in cls.COMPAT_ENGINES)
        )

    def draw(self, context):
        layout = self.layout

        ob = context.object
        psys = context.particle_system
        part = 0

        if ob:
            row = layout.row()

            row.template_list(
                "PARTICLE_UL_particle_systems", "particle_systems", ob, "particle_systems",
                ob.particle_systems, "active_index", rows=3,
            )

            col = row.column(align=True)
            col.operator("object.particle_system_add", icon='ADD', text="")
            col.operator("object.particle_system_remove", icon='REMOVE', text="")

            col.separator()

            col.menu("PARTICLE_MT_context_menu", icon='DOWNARROW_HLT', text="")

        if psys is None:
            part = particle_get_settings(context)

            if part is None:
                return

            layout.template_ID(context.space_data, "pin_id")

            if part.is_fluid:
                layout.label(text="Settings used for fluid")
                return

            layout.prop(part, "type", text="Type")

        elif not psys.settings:
            col.template_ID(psys, "settings", new="particle.new")
        else:
            part = psys.settings

            col = layout.column()

            if (part.is_fluid is False):
                row = col.row()
                row.enabled = particle_panel_enabled(context, psys)
                row.template_ID(psys, "settings", new="particle.new")

            if part.is_fluid:
                layout.label(text=rpt_("{:d} fluid particles for this frame").format(part.count), translate=False)
                return

            row = layout.row()
            row.enabled = particle_panel_enabled(context, psys)
            row.prop(part, "type", expand=True)

        if part:
            split = layout.split()
            if part.type == 'HAIR':
                if psys is not None and psys.is_edited:
                    split.operator("particle.edited_clear", text="Delete Edit")
                else:
                    row = split.row()
                    row.enabled = particle_panel_enabled(context, psys)
                    row.prop(part, "use_regrow_hair")
                    row.prop(part, "use_advanced_hair")

                if psys is not None and psys.is_edited:
                    if psys.is_global_hair:
                        row = layout.row(align=True)
                        row.operator("particle.connect_hair").all = False
                        row.operator("particle.connect_hair", text="Connect All").all = True
                    else:
                        row = layout.row(align=True)
                        row.operator("particle.disconnect_hair").all = False
                        row.operator("particle.disconnect_hair", text="Disconnect All").all = True
            elif psys is not None and part.type == 'REACTOR':
                split.enabled = particle_panel_enabled(context, psys)
                split.prop(psys, "reactor_target_object")
                split.prop(psys, "reactor_target_particle_system", text="Particle System")


class PARTICLE_PT_emission(ParticleButtonsPanel, Panel):
    bl_label = "Emission"
    bl_translation_context = i18n_contexts.id_particlesettings
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        psys = context.particle_system
        settings = particle_get_settings(context)

        if settings is None:
            return False
        if settings.is_fluid:
            return False
        if particle_panel_poll(PARTICLE_PT_emission, context):
            return psys is None or not context.particle_system.point_cache.use_external
        return False

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.use_property_split = True

        layout.enabled = particle_panel_enabled(context, psys) and (psys is None or not psys.has_multiple_caches)

        col = layout.column()
        col.active = part.emit_from == 'VERT' or part.distribution != 'GRID'
        col.prop(part, "count")

        if psys is not None:
            col.prop(psys, "seed")

        if part.type == 'HAIR':
            col.prop(part, "hair_length")
            col.prop(part, "hair_step")

        if part.type != 'HAIR':

            col = layout.column()

            sub = col.column(align=True)
            sub.prop(part, "frame_start", text="Frame Start")
            sub.prop(part, "frame_end", text="End")

            col.prop(part, "lifetime")
            col.prop(part, "lifetime_random", slider=True, text="Lifetime Randomness")


class PARTICLE_PT_emission_source(ParticleButtonsPanel, Panel):
    bl_label = "Source"
    bl_parent_id = "PARTICLE_PT_emission"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout

        part = particle_get_settings(context)

        layout.use_property_split = True

        col = layout.column()
        col.prop(part, "emit_from")
        col.prop(part, "use_modifier_stack")
        if part.emit_from in {'FACE', 'VOLUME'}:
            col.prop(part, "distribution")

        if part.emit_from == 'VERT':
            col.prop(part, "use_emit_random", text="Random Order")
        elif part.distribution == 'GRID':
            col.prop(part, "invert_grid")
            col.prop(part, "hexagonal_grid")
        else:
            col.prop(part, "use_emit_random", text="Random Order")
            col.prop(part, "use_even_distribution")

        if part.emit_from in {'FACE', 'VOLUME'}:

            if part.distribution == 'JIT':
                col.prop(part, "userjit", text="Particles/Face")
                col.prop(part, "jitter_factor", text="Jittering Amount", slider=True)
            elif part.distribution == 'GRID':
                col.prop(part, "grid_resolution")
                col.prop(part, "grid_random", text="Random", slider=True)


class PARTICLE_PT_hair_dynamics(ParticleButtonsPanel, Panel):
    bl_label = "Hair Dynamics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        psys = context.particle_system
        engine = context.engine
        if psys is None:
            return False
        if psys.settings is None:
            return False
        return psys.settings.type == 'HAIR' and (engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        psys = context.particle_system
        self.layout.prop(psys, "use_hair_dynamics", text="")

    def draw_header_preset(self, context):
        psys = context.particle_system

        if not psys.cloth:
            return

        layout = self.layout
        layout.enabled = psys.use_hair_dynamics and psys.point_cache.is_baked is False
        PARTICLE_PT_hair_dynamics_presets.draw_panel_header(layout)

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system

        if not psys.cloth:
            layout.label(text="Hair dynamics disabled")
            return

        cloth_md = psys.cloth
        cloth = cloth_md.settings
        result = cloth_md.solver_result

        layout.enabled = psys.use_hair_dynamics and psys.point_cache.is_baked is False

        layout.use_property_split = True

        layout.separator()

        col = layout.column()
        col.prop(cloth, "quality", text="Quality Steps", slider=True)

        layout.separator()

        col = layout.column()
        col.prop(cloth, "pin_stiffness", text="Pin Goal Strength")

        layout.separator()

        if result:
            box = layout.box()

            if not result.status:
                label = " "
                icon = 'NONE'
            elif result.status == {'SUCCESS'}:
                label = "Success"
                icon = 'NONE'
            elif result.status - {'SUCCESS'} == {'NO_CONVERGENCE'}:
                label = "No Convergence"
                icon = 'ERROR'
            else:
                label = "ERROR"
                icon = 'ERROR'
            box.label(text=label, icon=icon)
            box.label(
                text=rpt_("Iterations: {:d} .. {:d} (avg. {:d})").format(
                    result.min_iterations, result.max_iterations, result.avg_iterations,
                ),
                translate=False,
            )
            box.label(
                text=rpt_("Error: {:.5f} .. {:.5f} (avg. {:.5f})").format(
                    result.min_error, result.max_error, result.avg_error,
                ),
                translate=False,
            )


class PARTICLE_PT_hair_dynamics_collision(ParticleButtonsPanel, Panel):
    bl_label = "Collisions"
    bl_parent_id = "PARTICLE_PT_hair_dynamics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        return context.particle_system.cloth is not None

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        cloth_md = psys.cloth
        cloth_collision = cloth_md.collision_settings

        layout.enabled = psys.use_hair_dynamics and psys.point_cache.is_baked is False

        layout.use_property_split = True

        col = layout.column()
        col.prop(cloth_collision, "collision_quality", text="Quality")

        layout.separator()

        col = layout.column()
        col.prop(cloth_collision, "distance_min", slider=True, text="Distance")
        col.prop(cloth_collision, "impulse_clamp")
        col.prop(cloth_collision, "collection")


class PARTICLE_PT_hair_dynamics_structure(ParticleButtonsPanel, Panel):
    bl_label = "Structure"
    bl_parent_id = "PARTICLE_PT_hair_dynamics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        return context.particle_system.cloth is not None

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        cloth_md = psys.cloth
        cloth = cloth_md.settings

        layout.enabled = psys.use_hair_dynamics and psys.point_cache.is_baked is False

        layout.use_property_split = True

        col = layout.column()
        col.prop(cloth, "mass")
        sub = col.column(align=True)
        sub.prop(cloth, "bending_stiffness", text="Stiffness")
        sub.prop(psys.settings, "bending_random", text="Random")
        col.prop(cloth, "bending_damping", text="Damping")
        # XXX has no noticeable effect with stiff hair structure springs
        # col.prop(cloth, "spring_damping", text="Damping")


class PARTICLE_PT_hair_dynamics_volume(ParticleButtonsPanel, Panel):
    bl_label = "Volume"
    bl_parent_id = "PARTICLE_PT_hair_dynamics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        return context.particle_system.cloth is not None

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        cloth_md = psys.cloth
        cloth = cloth_md.settings

        layout.enabled = psys.use_hair_dynamics and psys.point_cache.is_baked is False

        layout.use_property_split = True

        col = layout.column()
        col.prop(cloth, "air_damping", text="Air Drag")
        col.prop(cloth, "internal_friction", slider=True)
        col.prop(cloth, "voxel_cell_size")

        col.separator()

        col.prop(cloth, "density_target", text="Density Target")
        col.prop(cloth, "density_strength", slider=True, text="Density Strength")


class PARTICLE_PT_cache(ParticleButtonsPanel, Panel):
    bl_label = "Cache"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        engine = context.engine
        if engine not in cls.COMPAT_ENGINES:
            return False
        psys = context.particle_system
        if psys is None:
            return False
        if psys.settings is None:
            return False
        if psys.settings.is_fluid:
            return False
        phystype = psys.settings.physics_type
        if phystype in {'NO', 'KEYED'}:
            return False
        return (
            psys.settings.type in {'EMITTER', 'REACTOR'} or (
                (psys.settings.type == 'HAIR') and
                (psys.use_hair_dynamics or psys.point_cache.is_baked)
            )
        )

    def draw(self, context):
        psys = context.particle_system

        point_cache_ui(self, psys.point_cache, True, 'HAIR' if (psys.settings.type == 'HAIR') else 'PSYS')


class PARTICLE_PT_velocity(ParticleButtonsPanel, Panel):
    bl_label = "Velocity"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if particle_panel_poll(PARTICLE_PT_velocity, context):
            psys = context.particle_system
            settings = particle_get_settings(context)

            if settings.type == 'HAIR' and not settings.use_advanced_hair:
                return False
            return settings.physics_type != 'BOIDS' and (psys is None or not psys.point_cache.use_external)
        else:
            return False

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.enabled = particle_panel_enabled(context, psys)
        layout.use_property_split = True

        col = layout.column()
        col.prop(part, "normal_factor")
        sub = col.column(align=True)
        sub.prop(part, "tangent_factor", text="Tangent")
        sub.prop(part, "tangent_phase", slider=True, text="Tangent Phase")

        col.separator()

        col.prop(part, "object_align_factor")

        col.separator()

        if part.emit_from == 'PARTICLE':
            col.prop(part, "particle_factor")
        else:
            col.prop(part, "object_factor", slider=True)
        col.prop(part, "factor_random", text="Randomize")

        # if part.type=='REACTOR':
        #     sub.prop(part, "reactor_factor")
        #     sub.prop(part, "reaction_shape", slider=True)


class PARTICLE_PT_rotation(ParticleButtonsPanel, Panel):
    bl_label = "Rotation"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if particle_panel_poll(PARTICLE_PT_rotation, context):
            psys = context.particle_system
            settings = particle_get_settings(context)

            if settings.type == 'HAIR' and not settings.use_advanced_hair:
                return False
            return settings.physics_type != 'BOIDS' and (psys is None or not psys.point_cache.use_external)
        else:
            return False

    def draw_header(self, context):
        psys = context.particle_system
        if psys:
            part = psys.settings
        else:
            part = context.space_data.pin_id

        layout = self.layout
        layout.prop(part, "use_rotations", text="")
        layout.enabled = particle_panel_enabled(context, psys)

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        if psys:
            part = psys.settings
        else:
            part = context.space_data.pin_id

        layout.enabled = particle_panel_enabled(context, psys) and part.use_rotations
        layout.use_property_split = True

        col = layout.column()

        col.prop(part, "rotation_mode")
        col.prop(part, "rotation_factor_random", slider=True, text="Randomize")

        col.separator()

        col.prop(part, "phase_factor", slider=True)
        col.prop(part, "phase_factor_random", text="Randomize Phase", slider=True)

        if part.type != 'HAIR':
            col.prop(part, "use_dynamic_rotation")


class PARTICLE_PT_rotation_angular_velocity(ParticleButtonsPanel, Panel):
    bl_label = "Angular Velocity"
    bl_parent_id = "PARTICLE_PT_rotation"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        if psys:
            part = psys.settings
        else:
            part = context.space_data.pin_id

        layout.enabled = particle_panel_enabled(context, psys) and part.use_rotations
        layout.use_property_split = True

        col = layout.column()

        col.prop(part, "angular_velocity_mode", text="Axis")
        sub = col.column(align=True)
        sub.active = part.angular_velocity_mode != 'NONE'
        sub.prop(part, "angular_velocity_factor", text="Amount")


class PARTICLE_PT_physics(ParticleButtonsPanel, Panel):
    bl_label = "Physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if particle_panel_poll(PARTICLE_PT_physics, context):
            psys = context.particle_system
            settings = particle_get_settings(context)

            if settings.type == 'HAIR' and not settings.use_advanced_hair:
                return False
            return psys is None or not psys.point_cache.use_external
        else:
            return False

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.enabled = particle_panel_enabled(context, psys)

        layout.prop(part, "physics_type")

        col = layout.column()
        if part.physics_type != 'NO':
            col = col.column()
            col.prop(part, "mass")
            col.prop(part, "use_multiply_size_mass", text="Multiply Mass with Size")

        if part.physics_type == 'FLUID':
            fluid = part.fluid

            col.separator()
            col.prop(fluid, "solver")
            col.prop(fluid, "stiffness", text="Stiffness")
            col.prop(fluid, "linear_viscosity", text="Viscosity")
            col.prop(fluid, "buoyancy", text="Buoyancy", slider=True)

        elif part.physics_type == 'KEYED':

            sub = col.column()
            sub.active = not psys or not psys.use_keyed_timing
            sub.prop(part, "keyed_loops", text="Loops")
            if psys:
                col.prop(psys, "use_keyed_timing", text="Use Timing")


class PARTICLE_PT_physics_fluid_advanced(ParticleButtonsPanel, Panel):
    bl_label = "Advanced"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type == 'FLUID'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        fluid = part.fluid

        col = layout.column()

        if fluid.solver == 'DDR':
            sub = col.column()
            sub.prop(fluid, "repulsion", slider=fluid.use_factor_repulsion)
            sub.prop(fluid, "use_factor_repulsion")

            sub.prop(fluid, "stiff_viscosity", slider=fluid.use_factor_stiff_viscosity)
            sub.prop(fluid, "use_factor_stiff_viscosity")

        sub = col.column()
        sub.prop(fluid, "fluid_radius", slider=fluid.use_factor_radius)
        sub.prop(fluid, "use_factor_radius")

        sub.prop(fluid, "rest_density", slider=fluid.use_factor_density)
        sub.prop(fluid, "use_factor_density")

        if fluid.solver == 'CLASSICAL':
            # With the classical solver, it is possible to calculate the
            # spacing between particles when the fluid is at rest. This
            # makes it easier to set stable initial conditions.
            particle_volume = part.mass / fluid.rest_density
            spacing = pow(particle_volume, 1.0 / 3.0)

            sub.label(text=iface_("Spacing: {:g}").format(spacing), translate=False)


class PARTICLE_PT_physics_fluid_springs(ParticleButtonsPanel, Panel):
    bl_label = "Springs"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        fluid = part.fluid
        return part.physics_type == 'FLUID' and fluid.solver == 'DDR'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        fluid = part.fluid

        col = layout.column()

        col.prop(fluid, "spring_force", text="Force")


class PARTICLE_PT_physics_fluid_springs_viscoelastic(ParticleButtonsPanel, Panel):
    bl_label = "Viscoelastic Springs"
    bl_parent_id = "PARTICLE_PT_physics_fluid_springs"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        fluid = part.fluid
        return part.physics_type == 'FLUID' and fluid.solver == 'DDR'

    def draw_header(self, context):
        part = particle_get_settings(context)
        fluid = part.fluid

        self.layout.prop(fluid, "use_viscoelastic_springs", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        fluid = part.fluid

        col = layout.column()
        col.active = fluid.use_viscoelastic_springs
        col.prop(fluid, "yield_ratio", slider=True)
        col.prop(fluid, "plasticity", slider=True)

        col.separator()

        col.prop(fluid, "use_initial_rest_length")
        col.prop(fluid, "spring_frames", text="Frames")


class PARTICLE_PT_physics_fluid_springs_advanced(ParticleButtonsPanel, Panel):
    bl_label = "Advanced"
    bl_parent_id = "PARTICLE_PT_physics_fluid_springs"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        fluid = part.fluid
        return part.physics_type == 'FLUID' and fluid.solver == 'DDR'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        fluid = part.fluid

        sub = layout.column()
        sub.prop(fluid, "rest_length", slider=fluid.use_factor_rest_length)
        sub.prop(fluid, "use_factor_rest_length")


class PARTICLE_PT_physics_boids_movement(ParticleButtonsPanel, Panel):
    bl_label = "Movement"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type == 'BOIDS'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        boids = part.boids

        col = layout.column()

        col.prop(boids, "use_flight")
        col.prop(boids, "use_land")
        col.prop(boids, "use_climb")

        col = layout.column()

        col.active = boids.use_flight
        sub = col.column()
        sub.prop(boids, "air_speed_max")
        sub.prop(boids, "air_speed_min", slider=True)
        col.prop(boids, "air_acc_max", slider=True)
        col.prop(boids, "air_ave_max", slider=True)
        col.prop(boids, "air_personal_space")
        row = col.row(align=True)
        row.active = (boids.use_land or boids.use_climb) and boids.use_flight
        row.prop(boids, "land_smooth")

        layout.separator()

        col = layout.column()
        col.active = boids.use_land or boids.use_climb
        col.prop(boids, "land_speed_max")
        col.prop(boids, "land_jump_speed")
        col.prop(boids, "land_acc_max", slider=True)
        col.prop(boids, "land_ave_max", slider=True)
        col.prop(boids, "land_personal_space")
        col.prop(boids, "land_stick_force")

        layout.separator()

        layout.prop(part, "collision_collection")


class PARTICLE_PT_physics_boids_battle(ParticleButtonsPanel, Panel):
    bl_label = "Battle"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type == 'BOIDS'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        boids = part.boids

        col = layout.column()

        col.prop(boids, "health")
        col.prop(boids, "strength")
        col.prop(boids, "aggression")
        col.prop(boids, "accuracy")
        col.prop(boids, "range")


class PARTICLE_PT_physics_boids_misc(ParticleButtonsPanel, Panel):
    bl_label = "Misc"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type == 'BOIDS'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)
        boids = part.boids

        col = layout.column()

        col.prop(boids, "bank", slider=True)
        col.prop(boids, "pitch", slider=True)
        col.prop(boids, "height", slider=True)


class PARTICLE_PT_physics_relations(ParticleButtonsPanel, Panel):
    bl_label = "Relations"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        psys = context.particle_system
        part = particle_get_settings(context)
        return psys and part.physics_type in {'KEYED', 'BOIDS'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        row = layout.row()
        row.template_list(
            "UI_UL_list", "particle_targets", psys, "targets",
            psys, "active_particle_target_index", rows=4,
        )

        col = row.column()
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator("particle.new_target", icon='ADD', text="")
        subsub.operator("particle.target_remove", icon='REMOVE', text="")
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator("particle.target_move_up", icon='TRIA_UP', text="")
        subsub.operator("particle.target_move_down", icon='TRIA_DOWN', text="")

        key = psys.active_particle_target

        if key:
            if part.physics_type == 'KEYED':
                col = layout.column()
                # doesn't work yet
                # col.alert = key.valid
                col.prop(key, "object")
                col.prop(key, "system", text="System")
                sub = col.column(align=True)
                sub.active = psys.use_keyed_timing
                sub.prop(key, "time")
                sub.prop(key, "duration")
            elif part.physics_type == 'BOIDS':
                sub = layout.column()
                # doesn't work yet
                # sub.alert = key.valid
                sub.prop(key, "object")
                sub.prop(key, "system", text="System")
                layout.prop(key, "alliance")


class PARTICLE_PT_physics_fluid_interaction(ParticleButtonsPanel, Panel):
    bl_label = "Fluid Interaction"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type == 'FLUID'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system

        row = layout.row()
        row.template_list(
            "UI_UL_list", "particle_targets", psys, "targets",
            psys, "active_particle_target_index", rows=4,
        )

        col = row.column()
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator("particle.new_target", icon='ADD', text="")
        subsub.operator("particle.target_remove", icon='REMOVE', text="")
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator("particle.target_move_up", icon='TRIA_UP', text="")
        subsub.operator("particle.target_move_down", icon='TRIA_DOWN', text="")

        key = psys.active_particle_target

        if key:
            sub = layout.column()
            # doesn't work yet
            # sub.alert = key.valid
            sub.prop(key, "object")
            sub.prop(key, "system", text="System")


class PARTICLE_PT_physics_deflection(ParticleButtonsPanel, Panel):
    bl_label = "Deflection"
    bl_parent_id = "PARTICLE_PT_physics"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type in {'NEWTON', 'FLUID'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.enabled = particle_panel_enabled(context, psys)

        col = layout.column()
        col.prop(part, "use_size_deflect")
        col.prop(part, "use_die_on_collision")

        col.prop(part, "collision_collection")


class PARTICLE_PT_physics_forces(ParticleButtonsPanel, Panel):
    bl_label = "Forces"
    bl_parent_id = "PARTICLE_PT_physics"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type in {'NEWTON', 'FLUID'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.enabled = particle_panel_enabled(context, psys)

        col = layout.column()

        col.prop(part, "brownian_factor")
        col.prop(part, "drag_factor", slider=True)
        col.prop(part, "damping", slider=True)


class PARTICLE_PT_physics_integration(ParticleButtonsPanel, Panel):
    bl_label = "Integration"
    bl_options = {'DEFAULT_CLOSED'}
    bl_parent_id = "PARTICLE_PT_physics"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.physics_type in {'NEWTON', 'FLUID'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.enabled = particle_panel_enabled(context, psys)

        col = layout.column()

        col.prop(part, "integrator")
        col.prop(part, "timestep")
        col.prop(part, "subframes")

        if part.physics_type == 'FLUID':
            col.prop(part, "use_adaptive_subframes", text="Adaptive")
            sub = col.row()
            sub.enabled = part.use_adaptive_subframes
            sub.prop(part, "courant_target", text="Threshold")


class PARTICLE_PT_boidbrain(ParticleButtonsPanel, Panel):
    bl_label = "Boid Brain"
    bl_options = {'DEFAULT_CLOSED'}
    bl_parent_id = "PARTICLE_PT_physics"

    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        psys = context.particle_system
        settings = particle_get_settings(context)
        engine = context.engine

        if settings is None:
            return False
        if psys is not None and psys.point_cache.use_external:
            return False
        return settings.physics_type == 'BOIDS' and engine in cls.COMPAT_ENGINES

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        boids = particle_get_settings(context).boids

        layout.enabled = particle_panel_enabled(context, context.particle_system)

        # Currently boids can only use the first state so these are commented out for now.
        # row = layout.row()
        # row.template_list("UI_UL_list", "particle_boids", boids, "states",
        #                   boids, "active_boid_state_index", compact="True")
        # col = row.row()
        # sub = col.row(align=True)
        # sub.operator("boid.state_add", icon='ADD', text="")
        # sub.operator("boid.state_del", icon='REMOVE', text="")
        # sub = row.row(align=True)
        # sub.operator("boid.state_move_up", icon='TRIA_UP', text="")
        # sub.operator("boid.state_move_down", icon='TRIA_DOWN', text="")

        state = boids.active_boid_state

        # layout.prop(state, "name", text="State name")

        row = layout.row()
        row.template_list(
            "UI_UL_list", "particle_boids_rules", state,
            "rules", state, "active_boid_rule_index", rows=4,
        )

        col = row.column()
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator_menu_enum("boid.rule_add", "type", icon='ADD', text="")
        subsub.operator("boid.rule_del", icon='REMOVE', text="")
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator("boid.rule_move_up", icon='TRIA_UP', text="")
        subsub.operator("boid.rule_move_down", icon='TRIA_DOWN', text="")

        layout.prop(state, "ruleset_type")
        if state.ruleset_type == 'FUZZY':
            layout.prop(state, "rule_fuzzy", slider=True)

        rule = state.active_boid_rule

        if rule:
            col = layout.column(align=True)
            col.prop(rule, "use_in_air")
            col.prop(rule, "use_on_land")

            col = layout.column()

            if rule.type == 'GOAL':
                col.prop(rule, "object")
                col.prop(rule, "use_predict")
            elif rule.type == 'AVOID':
                col.prop(rule, "object")
                col.prop(rule, "use_predict")
                col.prop(rule, "fear_factor")
            elif rule.type == 'FOLLOW_PATH':
                col.label(text="Not yet functional")
            elif rule.type == 'AVOID_COLLISION':
                col.prop(rule, "use_avoid")
                col.prop(rule, "use_avoid_collision")
                col.prop(rule, "look_ahead")
            elif rule.type == 'FOLLOW_LEADER':
                col.prop(rule, "object")
                col.prop(rule, "distance")
                col.prop(rule, "use_line")
                sub = col.row()
                sub.active = rule.use_line
                sub.prop(rule, "queue_count")
            elif rule.type == 'AVERAGE_SPEED':
                col.prop(rule, "speed", slider=True)
                col.prop(rule, "wander", slider=True)
                col.prop(rule, "level", slider=True)
            elif rule.type == 'FIGHT':
                col.prop(rule, "distance")
                col.prop(rule, "flee_distance")


class PARTICLE_PT_render(ParticleButtonsPanel, Panel):
    bl_label = "Render"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        settings = particle_get_settings(context)
        engine = context.engine
        if settings is None:
            return False

        return engine in cls.COMPAT_ENGINES

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.prop(part, "render_type", text="Render As")

        if (
                part.type == 'EMITTER' or
                part.type in {'FLIP', 'SPRAY', 'BUBBLE', 'FOAM', 'TRACER',
                              'SPRAYFOAM', 'SPRAYBUBBLE', 'FOAMBUBBLE', 'SPRAYFOAMBUBBLE'} or
                (part.render_type in {'OBJECT', 'COLLECTION'} and part.type == 'HAIR')
        ):
            if part.render_type != 'NONE':
                col = layout.column(align=True)
                col.prop(part, "particle_size", text="Scale")
                col.prop(part, "size_random", slider=True, text="Scale Randomness")

        if psys:
            col = layout.column()
            if part.render_type not in {'OBJECT', 'COLLECTION', 'NONE'}:
                # col.enabled = False
                col.prop(part, "material_slot", text="Material")
                col.prop(psys, "parent", text="Coordinate System")

        if context.object:
            layout.separator()
            layout.prop(context.object, "show_instancer_for_render", text="Show Emitter")


class PARTICLE_PT_render_extra(ParticleButtonsPanel, Panel):
    bl_label = "Extra"
    bl_parent_id = "PARTICLE_PT_render"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.render_type != 'NONE'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        col = layout.column()

        col = layout.column()
        col.prop(part, "use_parent_particles", text="Parent Particles")
        col.prop(part, "show_unborn", text="Unborn")
        col.prop(part, "use_dead", text="Dead")


class PARTICLE_PT_render_path(ParticleButtonsPanel, Panel):
    bl_label = "Path"
    bl_parent_id = "PARTICLE_PT_render"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.render_type == 'PATH'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        col = layout.column()
        col.prop(part, "use_hair_bspline")
        col.prop(part, "render_step", text="Steps")


class PARTICLE_PT_render_path_timing(ParticleButtonsPanel, Panel):
    bl_label = "Timing"
    bl_parent_id = "PARTICLE_PT_render"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.render_type == 'PATH'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        col = layout.column()

        col.prop(part, "use_absolute_path_time")

        if part.type == 'HAIR' or psys.point_cache.is_baked:
            col.prop(part, "path_start", text="Start", slider=not part.use_absolute_path_time)

        col.prop(part, "path_end", text="End", slider=not part.use_absolute_path_time)
        col.prop(part, "length_random", text="Random", slider=True)


class PARTICLE_PT_render_object(ParticleButtonsPanel, Panel):
    bl_label = "Object"
    bl_parent_id = "PARTICLE_PT_render"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.render_type == 'OBJECT'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        col = layout.column()

        col.prop(part, "instance_object", text="Instance Object")
        sub = col.column()
        sub.prop(part, "use_global_instance", text="Global Coordinates")
        sub.prop(part, "use_rotation_instance", text="Object Rotation")
        sub.prop(part, "use_scale_instance", text="Object Scale")


class PARTICLE_PT_render_collection(ParticleButtonsPanel, Panel):
    bl_label = "Collection"
    bl_parent_id = "PARTICLE_PT_render"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.render_type == 'COLLECTION'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        col = layout.column()

        col.prop(part, "instance_collection", text="Instance Collection")

        col.prop(part, "use_whole_collection")
        sub = col.column()
        sub.active = (part.use_whole_collection is False)
        sub.prop(part, "use_collection_pick_random")
        sub.prop(part, "use_global_instance", text="Global Coordinates")
        sub.prop(part, "use_rotation_instance", text="Object Rotation")
        sub.prop(part, "use_scale_instance", text="Object Scale")


class PARTICLE_PT_render_collection_use_count(ParticleButtonsPanel, Panel):
    bl_label = "Use Count"
    bl_parent_id = "PARTICLE_PT_render_collection"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.render_type == 'COLLECTION'

    def draw_header(self, context):
        layout = self.layout
        part = particle_get_settings(context)

        layout.active = not part.use_whole_collection

        layout.prop(part, "use_collection_count", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        col = layout.column()

        layout.active = part.use_collection_count and not part.use_whole_collection

        row = layout.row()
        row.template_list(
            "UI_UL_list", "particle_instance_weights", part, "instance_weights",
            part, "active_instanceweight_index",
        )

        col = row.column()
        sub = col.row()
        subsub = sub.column(align=True)
        subsub.operator("particle.dupliob_copy", icon='ADD', text="")
        subsub.operator("particle.dupliob_remove", icon='REMOVE', text="")
        subsub.operator("particle.dupliob_move_up", icon='TRIA_UP', text="")
        subsub.operator("particle.dupliob_move_down", icon='TRIA_DOWN', text="")
        subsub.separator()
        subsub.operator("particle.dupliob_refresh", icon='FILE_REFRESH', text="")

        weight = part.active_instanceweight
        if weight:
            row = layout.row()
            row.prop(weight, "count")


class PARTICLE_PT_draw(ParticleButtonsPanel, Panel):
    bl_label = "Viewport Display"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        settings = particle_get_settings(context)
        engine = context.engine
        if settings is None:
            return False
        return engine in cls.COMPAT_ENGINES

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.prop(part, "display_method", text="Display As")

        if part.display_method == 'NONE' or (part.render_type == 'NONE' and part.display_method == 'RENDER'):
            return

        path = (part.render_type == 'PATH' and part.display_method == 'RENDER') or part.display_method == 'PATH'

        layout.separator()

        col = layout.column()
        col.prop(part, "display_color", text="Color")
        if part.display_color in {'VELOCITY', 'ACCELERATION'}:
            col.prop(part, "color_maximum", text="Fade Distance")

        col = layout.column()

        if path:
            col.prop(part, "display_step", text="Strand Steps")
        col.prop(part, "display_percentage", slider=True, text="Amount", text_ctxt=i18n_contexts.countable)
        if part.display_method != 'RENDER' or part.render_type == 'HALO':
            col.prop(part, "display_size", text="Size")

        if part.display_percentage != 100 and psys is not None:
            if part.type == 'HAIR':
                if psys.use_hair_dynamics and psys.point_cache.is_baked is False:
                    layout.row().label(text="Display percentage makes dynamics inaccurate without baking")
            else:
                phystype = part.physics_type
                if phystype != 'NO' and phystype != 'KEYED' and psys.point_cache.is_baked is False:
                    layout.row().label(text="Display percentage makes dynamics inaccurate without baking")
        else:
            layout.separator()

        if context.object:
            layout.separator()
            layout.prop(context.object, "show_instancer_for_viewport", text="Show Emitter")


class PARTICLE_PT_children(ParticleButtonsPanel, Panel):
    bl_label = "Children"
    bl_translation_context = i18n_contexts.id_particlesettings
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        return particle_panel_poll(cls, context)

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        part = particle_get_settings(context)

        layout.row().prop(part, "child_type", expand=True)

        layout.use_property_split = True

        if part.child_type == 'NONE':
            return

        col = layout.column()

        sub = col.column(align=True)
        sub.prop(part, "child_percent", text="Display Amount")
        sub.prop(part, "rendered_child_count", text="Render Amount")

        col.separator()

        col.prop(part, "child_length", slider=True)
        col.prop(part, "child_length_threshold", slider=True)
        if psys:
            col.prop(psys, "child_seed", text="Seed")

        col.separator()

        if part.child_type == 'INTERPOLATED':
            col.prop(part, "virtual_parents", slider=True)
            col.prop(part, "create_long_hair_children")
        else:
            col.separator()
            sub = col.column(align=True)
            sub.prop(part, "child_size", text="Size")
            sub.prop(part, "child_size_random", text="Randomize Size", slider=True)

        if part.child_type == 'SIMPLE':
            col.separator()
            col.prop(part, "child_radius", text="Radius")
            col.prop(part, "child_roundness", text="Roundness", slider=True)


class PARTICLE_PT_children_parting(ParticleButtonsPanel, Panel):
    bl_label = "Parting"
    bl_parent_id = "PARTICLE_PT_children"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.child_type == 'INTERPOLATED'

    def draw(self, context):
        layout = self.layout

        part = particle_get_settings(context)
        is_virtual = part.virtual_parents > 0.0

        layout.use_property_split = True

        if is_virtual:
            layout.label(text="Parting not available with virtual parents")

        col = layout.column()
        col.active = not is_virtual
        col.prop(part, "child_parting_factor", text="Parting", slider=True)

        sub = col.column(align=True)
        sub.prop(part, "child_parting_min", text="Min")
        sub.prop(part, "child_parting_max", text="Max")


class PARTICLE_PT_children_clumping(ParticleButtonsPanel, Panel):
    bl_label = "Clumping"
    bl_parent_id = "PARTICLE_PT_children"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.child_type != 'NONE'

    def draw(self, context):
        layout = self.layout

        part = particle_get_settings(context)

        layout.use_property_split = True

        col = layout.column()

        sub = col.column()

        sub.prop(part, "use_clump_curve")
        if part.use_clump_curve:
            sub.template_curve_mapping(part, "clump_curve")
        else:
            sub.prop(part, "clump_factor", slider=True)
            sub.prop(part, "clump_shape", slider=True)

        if part.child_type == 'SIMPLE':
            col.prop(part, "twist")
            col.prop(part, "use_twist_curve")
            if part.use_twist_curve:
                col.template_curve_mapping(part, "twist_curve")


class PARTICLE_PT_children_clumping_noise(ParticleButtonsPanel, Panel):
    bl_label = "Clump Noise"
    bl_parent_id = "PARTICLE_PT_children_clumping"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw_header(self, context):

        part = particle_get_settings(context)

        self.layout.prop(part, "use_clump_noise", text="")

    def draw(self, context):
        layout = self.layout

        part = particle_get_settings(context)

        layout.use_property_split = True
        layout.enabled = part.use_clump_noise

        layout.prop(part, "clump_noise_size")


class PARTICLE_PT_children_roughness(ParticleButtonsPanel, Panel):
    bl_label = "Roughness"
    bl_translation_context = i18n_contexts.id_particlesettings
    bl_parent_id = "PARTICLE_PT_children"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.child_type != 'NONE'

    def draw(self, context):
        layout = self.layout

        part = particle_get_settings(context)

        layout.use_property_split = True

        col = layout.column()

        col.prop(part, "use_roughness_curve")
        if part.use_roughness_curve:
            sub = col.column()
            sub.template_curve_mapping(part, "roughness_curve")
            sub.prop(part, "roughness_1", text="Roughness", text_ctxt=i18n_contexts.id_particlesettings)
            sub.prop(part, "roughness_1_size", text="Size")
        else:
            sub = col.column(align=True)
            sub.prop(part, "roughness_1", text="Uniform")
            sub.prop(part, "roughness_1_size", text="Size")

            sub = col.column(align=True)
            sub.prop(part, "roughness_endpoint", text="Endpoint")
            sub.prop(part, "roughness_end_shape")

            sub = col.column(align=True)
            sub.prop(part, "roughness_2", text="Random")
            sub.prop(part, "roughness_2_size", text="Size")
            sub.prop(part, "roughness_2_threshold", slider=True)


class PARTICLE_PT_children_kink(ParticleButtonsPanel, Panel):
    bl_label = "Kink"
    bl_parent_id = "PARTICLE_PT_children"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        part = particle_get_settings(context)
        return part.child_type != 'NONE'

    def draw(self, context):
        layout = self.layout

        part = particle_get_settings(context)

        layout.use_property_split = True

        col = layout.column()

        col.prop(part, "kink", text="Kink Type")
        col = layout.column()
        col.active = part.kink != 'NO'

        if part.kink == 'SPIRAL':

            sub = col.column()
            sub.prop(part, "kink_amplitude", text="Amplitude")
            sub.prop(part, "kink_amplitude_random", text="Randomize Amplitude", slider=True)

            col.separator()

            sub = col.column()
            sub.prop(part, "kink_axis")
            sub.prop(part, "kink_axis_random", text="Randomize Axis", slider=True)

            col.separator()

            col.prop(part, "kink_frequency", text="Frequency")
            col.prop(part, "kink_shape", text="Shape", slider=True)
            col.prop(part, "kink_extra_steps", text="Steps")

        elif part.kink in {'CURL', 'RADIAL', 'WAVE', 'BRAID', 'WAVE'}:
            sub = col.column(align=True)
            sub.prop(part, "kink_amplitude")
            sub.prop(part, "kink_amplitude_clump", text="Clump", slider=True)
            col.prop(part, "kink_flat", slider=True)
            col.prop(part, "kink_frequency")
            col.prop(part, "kink_shape", slider=True)


class PARTICLE_PT_field_weights(ParticleButtonsPanel, Panel):
    bl_label = "Field Weights"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        return particle_panel_poll(cls, context)

    def draw(self, context):
        part = particle_get_settings(context)
        effector_weights_ui(self, part.effector_weights, 'PSYS')

        if part.type == 'HAIR':
            row = self.layout.row()
            row.prop(part.effector_weights, "apply_to_hair_growing")
            row.prop(part, "apply_effector_to_children")
            row = self.layout.row()
            row.prop(part, "effect_hair", slider=True)


class PARTICLE_PT_force_fields(ParticleButtonsPanel, Panel):
    bl_label = "Force Field Settings"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        col = layout.column()
        col.prop(part, "use_self_effect")
        col.prop(part, "effector_amount", text="Effector Amount")


class PARTICLE_PT_force_fields_type1(ParticleButtonsPanel, Panel):
    bl_label = "Type 1"
    bl_parent_id = "PARTICLE_PT_force_fields"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        if part.force_field_1:
            col = layout.column()
            col.prop(part.force_field_1, "type", text="Type 1")
            basic_force_field_settings_ui(self, part.force_field_1)


class PARTICLE_PT_force_fields_type2(ParticleButtonsPanel, Panel):
    bl_label = "Type 2"
    bl_parent_id = "PARTICLE_PT_force_fields"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        if part.force_field_2:
            col = layout.column()
            col.prop(part.force_field_2, "type", text="Type 2")
            basic_force_field_settings_ui(self, part.force_field_2)


class PARTICLE_PT_force_fields_type1_falloff(ParticleButtonsPanel, Panel):
    bl_label = "Falloff"
    bl_options = {'DEFAULT_CLOSED'}
    bl_parent_id = "PARTICLE_PT_force_fields_type1"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        if part.force_field_1:
            basic_force_field_falloff_ui(self, part.force_field_1)


class PARTICLE_PT_force_fields_type2_falloff(ParticleButtonsPanel, Panel):
    bl_label = "Falloff"
    bl_options = {'DEFAULT_CLOSED'}
    bl_parent_id = "PARTICLE_PT_force_fields_type2"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        part = particle_get_settings(context)

        if part.force_field_2:
            basic_force_field_falloff_ui(self, part.force_field_2)


class PARTICLE_PT_vertexgroups(ParticleButtonsPanel, Panel):
    bl_label = "Vertex Groups"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if context.particle_system is None:
            return False
        return particle_panel_poll(cls, context)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        ob = context.object
        psys = context.particle_system

        col = layout.column()
        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_density", ob, "vertex_groups", text="Density")
        row.prop(psys, "invert_vertex_group_density", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_length", ob, "vertex_groups", text="Length")
        row.prop(psys, "invert_vertex_group_length", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_clump", ob, "vertex_groups", text="Clump")
        row.prop(psys, "invert_vertex_group_clump", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_kink", ob, "vertex_groups", text="Kink")
        row.prop(psys, "invert_vertex_group_kink", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_roughness_1", ob, "vertex_groups", text="Roughness 1")
        row.prop(psys, "invert_vertex_group_roughness_1", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_roughness_2", ob, "vertex_groups", text="Roughness 2")
        row.prop(psys, "invert_vertex_group_roughness_2", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_roughness_end", ob, "vertex_groups", text="Roughness End")
        row.prop(psys, "invert_vertex_group_roughness_end", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        row = col.row(align=True)
        sub = row.row(align=True)
        sub.use_property_decorate = False
        sub.prop_search(psys, "vertex_group_twist", ob, "vertex_groups", text="Twist")
        row.prop(psys, "invert_vertex_group_twist", text="", toggle=True, icon='ARROW_LEFTRIGHT')

        # Commented out vertex groups don't work and are still waiting for better implementation
        # row = layout.row()
        # row.prop_search(psys, "vertex_group_velocity", ob, "vertex_groups", text="Velocity")
        # row.prop(psys, "invert_vertex_group_velocity", text="")

        # row = layout.row()
        # row.prop_search(psys, "vertex_group_size", ob, "vertex_groups", text="Size")
        # row.prop(psys, "invert_vertex_group_size", text="")

        # row = layout.row()
        # row.prop_search(psys, "vertex_group_tangent", ob, "vertex_groups", text="Tangent")
        # row.prop(psys, "invert_vertex_group_tangent", text="")

        # row = layout.row()
        # row.prop_search(psys, "vertex_group_rotation", ob, "vertex_groups", text="Rotation")
        # row.prop(psys, "invert_vertex_group_rotation", text="")

        # row = layout.row()
        # row.prop_search(psys, "vertex_group_field", ob, "vertex_groups", text="Field")
        # row.prop(psys, "invert_vertex_group_field", text="")


class PARTICLE_PT_textures(ParticleButtonsPanel, Panel):
    bl_label = "Textures"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if context.particle_system is None:
            return False
        return particle_panel_poll(cls, context)

    def draw(self, context):
        layout = self.layout

        psys = context.particle_system
        part = psys.settings

        row = layout.row()
        row.template_list("TEXTURE_UL_texslots", "", part, "texture_slots", part, "active_texture_index", rows=2)

        col = row.column(align=True)
        col.operator("texture.slot_move", text="", icon='TRIA_UP').type = 'UP'
        col.operator("texture.slot_move", text="", icon='TRIA_DOWN').type = 'DOWN'
        col.menu("TEXTURE_MT_context_menu", icon='DOWNARROW_HLT', text="")

        if not part.active_texture:
            layout.template_ID(part, "active_texture", new="texture.new")
        else:
            slot = part.texture_slots[part.active_texture_index]
            layout.template_ID(slot, "texture", new="texture.new")


class PARTICLE_PT_hair_shape(ParticleButtonsPanel, Panel):
    bl_label = "Hair Shape"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        psys = context.particle_system
        if psys is None:
            return False
        return particle_panel_poll(cls, context) and psys.settings.type == 'HAIR'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        psys = context.particle_system
        part = psys.settings

        layout.prop(part, "shape", text="Strand Shape")

        col = layout.column(align=True)
        col.prop(part, "root_radius", text="Diameter Root")
        col.prop(part, "tip_radius", text="Tip")

        col = layout.column()
        col.prop(part, "radius_scale")
        col.prop(part, "use_close_tip")


class PARTICLE_PT_animation(ParticleButtonsPanel, PropertiesAnimationMixin, PropertyPanel, Panel):
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def _animated_id(cls, context):
        psys = getattr(context, "particle_system", None)
        return psys and psys.settings


class PARTICLE_PT_custom_props(ParticleButtonsPanel, PropertyPanel, Panel):
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE',
        'BLENDER_WORKBENCH',
    }
    _context_path = "particle_system.settings"
    _property_type = bpy.types.ParticleSettings


classes = (
    PARTICLE_MT_context_menu,
    PARTICLE_PT_hair_dynamics_presets,
    PARTICLE_UL_particle_systems,
    PARTICLE_PT_context_particles,
    PARTICLE_PT_emission,
    PARTICLE_PT_emission_source,
    PARTICLE_PT_hair_dynamics,
    PARTICLE_PT_hair_dynamics_collision,
    PARTICLE_PT_hair_dynamics_structure,
    PARTICLE_PT_hair_dynamics_volume,
    PARTICLE_PT_cache,
    PARTICLE_PT_velocity,
    PARTICLE_PT_rotation,
    PARTICLE_PT_rotation_angular_velocity,
    PARTICLE_PT_physics,
    PARTICLE_PT_physics_boids_movement,
    PARTICLE_PT_physics_boids_battle,
    PARTICLE_PT_physics_boids_misc,
    PARTICLE_PT_physics_forces,
    PARTICLE_PT_physics_deflection,
    PARTICLE_PT_physics_integration,
    PARTICLE_PT_physics_relations,
    PARTICLE_PT_physics_fluid_springs,
    PARTICLE_PT_physics_fluid_springs_viscoelastic,
    PARTICLE_PT_physics_fluid_springs_advanced,
    PARTICLE_PT_physics_fluid_advanced,
    PARTICLE_PT_physics_fluid_interaction,
    PARTICLE_PT_boidbrain,
    PARTICLE_PT_render,
    PARTICLE_PT_render_path,
    PARTICLE_PT_render_path_timing,
    PARTICLE_PT_render_object,
    PARTICLE_PT_render_collection,
    PARTICLE_PT_render_collection_use_count,
    PARTICLE_PT_render_extra,
    PARTICLE_PT_draw,
    PARTICLE_PT_children,
    PARTICLE_PT_children_parting,
    PARTICLE_PT_children_clumping,
    PARTICLE_PT_children_clumping_noise,
    PARTICLE_PT_children_roughness,
    PARTICLE_PT_children_kink,
    PARTICLE_PT_hair_shape,
    PARTICLE_PT_field_weights,
    PARTICLE_PT_force_fields,
    PARTICLE_PT_force_fields_type1,
    PARTICLE_PT_force_fields_type1_falloff,
    PARTICLE_PT_force_fields_type2,
    PARTICLE_PT_force_fields_type2_falloff,
    PARTICLE_PT_vertexgroups,
    PARTICLE_PT_textures,
    PARTICLE_PT_animation,
    PARTICLE_PT_custom_props,
)

if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

from typing import TYPE_CHECKING, List, Sequence, Optional

import bpy
from bpy.props import StringProperty
from bpy.app.translations import pgettext_rpt as rpt_

import os
import re
import importlib
import traceback
from zipfile import ZipFile
from shutil import rmtree

from . import feature_sets

if TYPE_CHECKING:
    from . import RigifyFeatureSets


DEFAULT_NAME = 'rigify'

# noinspection PyProtectedMember
INSTALL_PATH = feature_sets._install_path()
NAME_PREFIX = feature_sets.__name__.split('.')

# noinspection SpellCheckingInspection
PROMOTED_FEATURE_SETS = [
    {
        "name": "Experimental Rigs by Alexander Gavrilov",
        "author": "Alexander Gavrilov",
        "description":
            "Experimental and/or niche rigs made by a Rigify maintainer.\n"
            "Includes a BlenRig-like spine, Body IK (knee & elbow IK), jiggles, skin transforms, etc.",
        "link": "https://github.com/angavrilov/angavrilov-rigs",
    },
    {
        "name": "Cessen's Rigify Extensions",
        "author": "Nathan Vegdahl",
        "description": "Collection of original legacy Rigify rigs minimally ported to the modern Rigify",
        "warning": "This feature set is maintained at the bare minimal level",
        "link": "https://github.com/cessen/cessen_rigify_ext",
    }
]


def get_install_path(*, create=False):
    if not os.path.exists(INSTALL_PATH):
        if create:
            os.makedirs(INSTALL_PATH, exist_ok=True)
        else:
            return None

    return INSTALL_PATH


def get_installed_modules_names() -> List[str]:
    """Return a list of module names of all feature sets in the file system."""
    features_path = get_install_path()
    if not features_path:
        return []

    sets = []

    for fs in os.listdir(features_path):
        if fs and fs[0] != '.' and fs != DEFAULT_NAME:
            fs_path = os.path.join(features_path, fs)
            if os.path.isdir(fs_path):
                sets.append(fs)

    return sets


def get_prefs_feature_sets() -> Sequence['RigifyFeatureSets']:
    from . import RigifyPreferences
    return RigifyPreferences.get_instance().rigify_feature_sets


def get_enabled_modules_names() -> List[str]:
    """Return a list of module names of all enabled feature sets."""
    installed_module_names = get_installed_modules_names()
    rigify_feature_sets = get_prefs_feature_sets()

    enabled_module_names = {fs.module_name for fs in rigify_feature_sets if fs.enabled}

    return [name for name in installed_module_names if name in enabled_module_names]


def find_module_name_by_link(link: str) -> str | None:
    """Returns the name of the feature set module that is associated with the specified url."""
    if not link:
        return None

    for fs in get_prefs_feature_sets():
        if fs.link == link:
            return fs.module_name or None

    return None


def mark_feature_set_exception(module_name: str):
    if module_name:
        for fs in get_prefs_feature_sets():
            if fs.module_name == module_name:
                fs.has_exceptions = True


def get_module(feature_set: str):
    return importlib.import_module('.'.join([*NAME_PREFIX, feature_set]))


def get_module_safe(feature_set: str):
    # noinspection PyBroadException
    try:
        return get_module(feature_set)
    except:  # noqa: E722
        return None


def get_module_by_link_safe(link: str):
    if module_name := find_module_name_by_link(link):
        return get_module_safe(module_name)


def get_dir_path(feature_set: str, *extra_items: list[str]):
    base_dir = os.path.join(INSTALL_PATH, feature_set, *extra_items)
    base_path = [*NAME_PREFIX, feature_set, *extra_items]
    return base_dir, base_path


def get_info_dict(feature_set: str):
    module = get_module_safe(feature_set)

    if module and hasattr(module, 'rigify_info'):
        data = module.rigify_info
        if isinstance(data, dict):
            return data

    return {}


def call_function_safe(module_name: str, func_name: str,
                       args: Optional[list] = None, kwargs: Optional[dict] = None,
                       mark_error=False):
    module = get_module_safe(module_name)

    if module:
        func = getattr(module, func_name, None)

        if callable(func):
            # noinspection PyBroadException
            try:
                return func(*(args or []), **(kwargs or {}))
            except Exception:
                print(f"Rigify Error: Could not call function '{func_name}' of feature set "
                      f"'{module_name}': exception occurred.\n")
                traceback.print_exc()
                print("")

                if mark_error:
                    mark_feature_set_exception(module_name)

    return None


def call_register_function(feature_set: str, do_register: bool):
    call_function_safe(feature_set, 'register' if do_register else 'unregister', mark_error=do_register)


def get_ui_name(feature_set: str):
    # Try to get user-defined name
    info = get_info_dict(feature_set)
    if 'name' in info:
        return info['name']

    # Default name based on directory
    name = re.sub(r'[_.-]', ' ', feature_set)
    name = re.sub(r'(?<=\d) (?=\d)', '.', name)
    return name.title()


def feature_set_items(_scene, _context):
    """Get items for the Feature Set EnumProperty"""
    items = [
        ('all', 'All', 'All installed feature sets and rigs bundled with Rigify'),
        ('rigify', 'Rigify Built-in', 'Rigs bundled with Rigify'),
    ]

    for fs in get_enabled_modules_names():
        ui_name = get_ui_name(fs)
        items.append((fs, ui_name, ui_name))

    return items


def verify_feature_set_archive(zipfile):
    """Verify that the zip file contains one root directory, and some required files."""
    dirname = None
    init_found = False
    data_found = False

    for name in zipfile.namelist():
        parts = re.split(r'[/\\]', name)

        if dirname is None:
            dirname = parts[0]
        elif dirname != parts[0]:
            dirname = None
            break

        if len(parts) == 2 and parts[1] == '__init__.py':
            init_found = True

        if len(parts) > 2 and parts[1] in {'rigs', 'metarigs'} and parts[-1] == '__init__.py':
            data_found = True

    return dirname, init_found, data_found


# noinspection PyPep8Naming
class DATA_OT_rigify_add_feature_set(bpy.types.Operator):
    bl_idname = "wm.rigify_add_feature_set"
    bl_label = "Add External Feature Set"
    bl_description = "Add external feature set (rigs, metarigs, ui templates)"
    bl_options = {"REGISTER", "UNDO", "INTERNAL"}

    filter_glob: StringProperty(default="*.zip", options={'HIDDEN'})
    filepath: StringProperty(maxlen=1024, subtype='FILE_PATH', options={'HIDDEN', 'SKIP_SAVE'})

    @classmethod
    def poll(cls, context):
        return True

    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}

    def execute(self, context):
        from . import RigifyPreferences
        addon_prefs = RigifyPreferences.get_instance()

        rigify_config_path = get_install_path(create=True)

        with ZipFile(bpy.path.abspath(self.filepath), 'r') as zip_archive:
            base_dirname, init_found, data_found = verify_feature_set_archive(zip_archive)

            if not base_dirname:
                self.report({'ERROR'}, "The feature set archive must contain one base directory")
                return {'CANCELLED'}

            # Patch up some invalid characters to allow using 'Download ZIP' on GitHub.
            fixed_dirname = re.sub(r'[.-]', '_', base_dirname)

            if not re.fullmatch(r'[a-zA-Z][a-zA-Z_0-9]*', fixed_dirname):
                message = rpt_("The feature set archive base directory name is not a valid "
                               "identifier: '{:s}'").format(base_dirname)
                self.report({'ERROR'}, message)
                return {'CANCELLED'}

            if fixed_dirname == DEFAULT_NAME:
                message = rpt_("The name '{:s}' is not allowed for feature sets").format(DEFAULT_NAME)
                self.report({'ERROR'}, message)
                return {'CANCELLED'}

            if not init_found or not data_found:
                self.report(
                    {'ERROR'},
                    "The feature set archive has no rigs or metarigs, or is missing __init__.py")
                return {'CANCELLED'}

            base_dir = os.path.join(rigify_config_path, base_dirname)
            fixed_dir = os.path.join(rigify_config_path, fixed_dirname)

            for path, name in [(base_dir, base_dirname), (fixed_dir, fixed_dirname)]:
                if os.path.exists(path):
                    message = rpt_("Feature set directory already exists: '{:s}'").format(name)
                    self.report({'ERROR'}, message)
                    return {'CANCELLED'}

            # Unpack the validated archive and fix the directory name if necessary
            zip_archive.extractall(rigify_config_path)

            if base_dir != fixed_dir:
                os.rename(base_dir, fixed_dir)

            # Call the register callback of the new set
            addon_prefs.refresh_installed_feature_sets()

            call_register_function(fixed_dirname, True)

            addon_prefs.update_external_rigs()

            # Select the new entry
            for i, fs in enumerate(addon_prefs.rigify_feature_sets):
                if fs.module_name == fixed_dirname:
                    addon_prefs.active_feature_set_index = i
                    break

        return {'FINISHED'}


# noinspection PyPep8Naming
class DATA_OT_rigify_remove_feature_set(bpy.types.Operator):
    bl_idname = "wm.rigify_remove_feature_set"
    bl_label = "Remove External Feature Set"
    bl_description = "Remove external feature set (rigs, metarigs, ui templates)"
    bl_options = {"REGISTER", "UNDO", "INTERNAL"}

    @classmethod
    def poll(cls, context):
        return True

    def invoke(self, context, event):
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        from . import RigifyPreferences
        addon_prefs = RigifyPreferences.get_instance()
        feature_set_list = addon_prefs.rigify_feature_sets
        active_idx = addon_prefs.active_feature_set_index
        active_fs: 'RigifyFeatureSets' = feature_set_list[active_idx]

        # Call the 'unregister' callback of the set being removed.
        if active_fs.enabled:
            call_register_function(active_fs.module_name, do_register=False)

        # Remove the feature set's folder from the file system.
        rigify_config_path = get_install_path()
        if rigify_config_path:
            set_path = os.path.join(rigify_config_path, active_fs.module_name)
            if os.path.exists(set_path):
                rmtree(set_path)

        # Remove the feature set's entry from the addon preferences.
        feature_set_list.remove(active_idx)

        # Remove the feature set's entries from the metarigs and rig types.
        addon_prefs.refresh_installed_feature_sets()
        addon_prefs.update_external_rigs()

        # Update active index.
        addon_prefs.active_feature_set_index -= 1

        return {'FINISHED'}


def register():
    bpy.utils.register_class(DATA_OT_rigify_add_feature_set)
    bpy.utils.register_class(DATA_OT_rigify_remove_feature_set)


def unregister():
    bpy.utils.unregister_class(DATA_OT_rigify_add_feature_set)
    bpy.utils.unregister_class(DATA_OT_rigify_remove_feature_set)



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List, Dict, Any


class Extension:
    """Container for extensions. Allows to specify requiredness"""
    extension = True  # class method used to check Extension class at traversal (after reloading script, isinstance is not working)

    def __init__(self, name: str, extension: Dict[str, Any], required: bool = True):
        self.name = name
        self.extension = extension
        self.required = required


class ChildOfRootExtension(Extension):
    """Container object for extensions that should be appended to the root extensions"""

    def __init__(self, path: List[str], name: str, extension: Dict[str, Any], required: bool = True):
        """
        Wrap a local extension entity into an object that will later be inserted into a root extension and converted
        to a reference.
        :param path: The path of the extension object in the root extension. E.g. ['lights'] for
        KHR_lights_punctual. Must be a path to a list in the extensions dict.
        :param extension: The data that should be placed into the extension list
        """
        self.path = path
        super().__init__(name, extension, required)



# SPDX-FileCopyrightText: 2018-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.app.handlers import persistent


def update_factory_startup_screens():
    screen = bpy.data.screens["Video Editing"]
    for area in screen.areas:
        if area.type == 'FILE_BROWSER':
            space = area.spaces.active
            params = space.params
            params.use_filter_folder = True


def update_factory_startup_ffmpeg_preset():
    from bpy import context

    preset = "H264_in_MP4"
    preset_filepath = bpy.utils.preset_find(preset, preset_path="ffmpeg")
    if not preset_filepath:
        print("Preset {!r} not found".format(preset))

    for scene in bpy.data.scenes:
        render = scene.render
        render.image_settings.media_type = 'VIDEO'
        render.image_settings.file_format = 'FFMPEG'

        if preset_filepath:
            with context.temp_override(scene=scene):
                bpy.ops.script.python_file_run(filepath=preset_filepath)

        render.ffmpeg.audio_codec = 'AAC'
        render.ffmpeg.audio_bitrate = 256


@persistent
def load_handler(_):
    update_factory_startup_screens()
    if bpy.app.build_options.codec_ffmpeg:
        update_factory_startup_ffmpeg_preset()


def register():
    bpy.app.handlers.load_factory_startup_post.append(load_handler)


def unregister():
    bpy.app.handlers.load_factory_startup_post.remove(load_handler)



# SPDX-FileCopyrightText: 2011-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from ..chain_rigs import SimpleChainRig
from ...base_rig import stage


class Rig(SimpleChainRig):
    """ A "copy_chain" rig.  All it does is duplicate the original bone chain
        and constrain it.
        This is a control and deformation rig.
    """

    make_controls: bool
    make_deforms: bool

    def initialize(self):
        super().initialize()

        """ Gather and validate data about the rig.
        """
        self.make_controls = self.params.make_controls
        self.make_deforms = self.params.make_deforms

    ##############################
    # Control chain

    @stage.generate_bones
    def make_control_chain(self):
        if self.make_controls:
            super().make_control_chain()

    @stage.parent_bones
    def parent_control_chain(self):
        if self.make_controls:
            super().parent_control_chain()

    @stage.configure_bones
    def configure_control_chain(self):
        if self.make_controls:
            super().configure_control_chain()

    @stage.generate_widgets
    def make_control_widgets(self):
        if self.make_controls:
            super().make_control_widgets()

    ##############################
    # ORG chain

    @stage.rig_bones
    def rig_org_chain(self):
        if self.make_controls:
            super().rig_org_chain()

    ##############################
    # Deform chain

    @stage.generate_bones
    def make_deform_chain(self):
        if self.make_deforms:
            super().make_deform_chain()

    @stage.parent_bones
    def parent_deform_chain(self):
        if self.make_deforms:
            super().parent_deform_chain()

    @stage.rig_bones
    def rig_deform_chain(self):
        if self.make_deforms:
            super().rig_deform_chain()

    ##############################
    # Parameter UI

    @classmethod
    def add_parameters(cls, params):
        """ Add the parameters of this rig type to the
            RigifyParameters PropertyGroup
        """
        params.make_controls = bpy.props.BoolProperty(
            name="Controls", default=True, description="Create control bones for the copy")
        params.make_deforms = bpy.props.BoolProperty(
            name="Deform", default=True, description="Create deform bones for the copy")

    @classmethod
    def parameters_ui(cls, layout, params):
        """ Create the ui for the rig parameters.
        """
        r = layout.row()
        r.prop(params, "make_controls")
        r = layout.row()
        r.prop(params, "make_deforms")


def create_sample(obj):
    """ Create a sample metarig for this rig type.
    """
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('bone.01')
    bone.head[:] = 0.0000, 0.0000, 0.0000
    bone.tail[:] = 0.0000, 0.0000, 0.3333
    bone.roll = 0.0000
    bone.use_connect = False
    bones['bone.01'] = bone.name
    bone = arm.edit_bones.new('bone.02')
    bone.head[:] = 0.0000, 0.0000, 0.3333
    bone.tail[:] = 0.0000, 0.0000, 0.6667
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['bone.01']]
    bones['bone.02'] = bone.name
    bone = arm.edit_bones.new('bone.03')
    bone.head[:] = 0.0000, 0.0000, 0.6667
    bone.tail[:] = 0.0000, 0.0000, 1.0000
    bone.roll = 3.1416
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['bone.02']]
    bones['bone.03'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['bone.01']]
    pbone.rigify_type = 'basic.copy_chain'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['bone.02']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['bone.03']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone
        if bcoll := arm.collections.active:
            bcoll.assign(bone)

    return bones



# SPDX-FileCopyrightText: 2020-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Initialization script for 2D Animation template

import bpy
from bpy.app.handlers import persistent


def update_factory_startup_screens():
    # 2D Animation.
    screen = bpy.data.screens["2D Animation"]
    for area in screen.areas:
        if area.type == 'PROPERTIES':
            # Set Tool settings as default in properties panel.
            space = area.spaces.active
            space.context = 'TOOL'
        elif area.type == 'DOPESHEET_EDITOR':
            # Open sidebar in Dope-sheet.
            space = area.spaces.active
            space.show_region_ui = True

    # 2D Full Canvas.
    screen = bpy.data.screens["2D Full Canvas"]
    for area in screen.areas:
        if area.type == 'VIEW_3D':
            space = area.spaces.active
            space.shading.type = 'MATERIAL'
            space.shading.use_scene_world = True


def update_factory_startup_scenes():
    for scene in bpy.data.scenes:
        scene.tool_settings.use_keyframe_insert_auto = True
        scene.tool_settings.gpencil_sculpt.use_scale_thickness = True


def update_factory_startup_grease_pencils():
    for grease_pencil in bpy.data.grease_pencils:
        grease_pencil.onion_keyframe_type = 'ALL'


@persistent
def load_handler(_):
    update_factory_startup_screens()
    update_factory_startup_scenes()
    update_factory_startup_grease_pencils()


def register():
    bpy.app.handlers.load_factory_startup_post.append(load_handler)


def unregister():
    bpy.app.handlers.load_factory_startup_post.remove(load_handler)



# SPDX-FileCopyrightText: 2018-2021 The glTF-Blender-IO authors
#
# SPDX-License-Identifier: Apache-2.0


from .action import gather_actions_animations
from .scene_animation import gather_scene_animations
from .tracks import gather_tracks_animations


def gather_animations(export_settings):

    # Reinit stored data
    export_settings['ranges'] = {}
    export_settings['slide'] = {}

    if export_settings['gltf_animation_mode'] in ["ACTIVE_ACTIONS", "ACTIONS", "BROADCAST"]:
        return gather_actions_animations(export_settings)
    elif export_settings['gltf_animation_mode'] == "SCENE":
        return gather_scene_animations(export_settings)
    elif export_settings['gltf_animation_mode'] == "NLA_TRACKS":
        return gather_tracks_animations(export_settings)



# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import (
    Panel,
)
from bpy.app.translations import (
    contexts as i18n_contexts,
    pgettext_rpt as rpt_,
)
from rna_prop_ui import PropertyPanel


class StripButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "strip"

    @classmethod
    def poll(cls, context):
        return context.active_strip is not None


class StripColorTagPicker:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "none"  # Used as popover

    @classmethod
    def poll(cls, context):
        return context.active_strip is not None


class STRIP_PT_color_tag_picker(StripColorTagPicker, Panel):
    bl_label = "Color Tag"
    bl_options = {'HIDE_HEADER'}

    def draw(self, _context):
        layout = self.layout

        row = layout.row(align=True)
        row.operator("sequencer.strip_color_tag_set", icon='X').color = 'NONE'
        for i in range(1, 10):
            icon = 'STRIP_COLOR_{:02d}'.format(i)
            row.operator("sequencer.strip_color_tag_set", icon=icon).color = 'COLOR_{:02d}'.format(i)


class STRIP_PT_strip(StripButtonsPanel, Panel):
    bl_label = ""
    bl_options = {'HIDE_HEADER'}

    def draw(self, context):
        layout = self.layout
        strip = context.active_strip
        strip_type = strip.type

        if strip_type in {
                'ADD', 'SUBTRACT', 'ALPHA_OVER', 'ALPHA_UNDER', 'MULTIPLY',
                'GLOW', 'SPEED', 'MULTICAM', 'GAUSSIAN_BLUR', 'COLORMIX',
        }:
            icon_header = 'SHADERFX'
        elif strip_type in {
                'CROSS', 'GAMMA_CROSS', 'WIPE',
        }:
            icon_header = 'ARROW_LEFTRIGHT'
        elif strip_type == 'SCENE':
            icon_header = 'SCENE_DATA'
        elif strip_type == 'MOVIECLIP':
            icon_header = 'TRACKER'
        elif strip_type == 'MASK':
            icon_header = 'MOD_MASK'
        elif strip_type == 'MOVIE':
            icon_header = 'FILE_MOVIE'
        elif strip_type == 'SOUND':
            icon_header = 'FILE_SOUND'
        elif strip_type == 'IMAGE':
            icon_header = 'FILE_IMAGE'
        elif strip_type == 'COLOR':
            icon_header = 'COLOR'
        elif strip_type == 'TEXT':
            icon_header = 'FONT_DATA'
        elif strip_type == 'ADJUSTMENT':
            icon_header = 'COLOR'
        elif strip_type == 'META':
            icon_header = 'SEQ_STRIP_META'
        else:
            icon_header = 'SEQ_SEQUENCER'

        row = layout.row(align=True)
        row.use_property_decorate = False
        row.label(text="", icon=icon_header)
        row.separator()
        row.prop(strip, "name", text="")

        sub = row.row(align=True)
        if strip.color_tag == 'NONE':
            sub.popover(panel="STRIP_PT_color_tag_picker", text="", icon='COLOR')
        else:
            icon = 'STRIP_' + strip.color_tag
            sub.popover(panel="STRIP_PT_color_tag_picker", text="", icon=icon)

        row.separator()
        row.prop(strip, "mute", toggle=True, icon_only=True, emboss=False)


class STRIP_PT_adjust_crop(StripButtonsPanel, Panel):
    bl_label = "Crop"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type != 'SOUND'

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True
        layout.active = not strip.mute

        col = layout.column(align=True)
        col.prop(strip.crop, "min_x")
        col.prop(strip.crop, "max_x")
        col.prop(strip.crop, "max_y")
        col.prop(strip.crop, "min_y")


class STRIP_PT_effect(StripButtonsPanel, Panel):
    bl_label = "Effect Strip"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type in {
            'ADD',
            'SUBTRACT',
            'ALPHA_OVER',
            'ALPHA_UNDER',
            'CROSS',
            'GAMMA_CROSS',
            'MULTIPLY',
            'WIPE',
            'GLOW',
            'COLOR',
            'SPEED',
            'MULTICAM',
            'GAUSSIAN_BLUR',
            'TEXT',
            'COLORMIX',
        }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        strip = context.active_strip

        layout.active = not strip.mute

        if strip.input_count > 0:
            col = layout.column()
            row = col.row()
            row.prop(strip, "input_1")

            if strip.input_count > 1:
                row.operator("sequencer.swap_inputs", text="", icon='SORT_ASC')
                row = col.row()
                row.prop(strip, "input_2")
                row.operator("sequencer.swap_inputs", text="", icon='SORT_DESC')

        strip_type = strip.type

        if strip_type == 'COLOR':
            layout.template_color_picker(strip, "color", value_slider=True, cubic=True)
            layout.prop(strip, "color", text="")

        elif strip_type == 'WIPE':
            col = layout.column()
            col.prop(strip, "transition_type")
            col.alignment = 'RIGHT'
            col.row().prop(strip, "direction", expand=True)

            col = layout.column()
            col.prop(strip, "blur_width", slider=True)
            if strip.transition_type in {'SINGLE', 'DOUBLE'}:
                col.prop(strip, "angle")

        elif strip_type == 'GLOW':
            flow = layout.column_flow()
            flow.prop(strip, "threshold", slider=True)
            flow.prop(strip, "clamp", slider=True)
            flow.prop(strip, "boost_factor")
            flow.prop(strip, "blur_radius")
            flow.prop(strip, "quality", slider=True)
            flow.prop(strip, "use_only_boost")

        elif strip_type == 'SPEED':
            col = layout.column(align=True)
            col.prop(strip, "speed_control", text="Speed Control")
            if strip.speed_control == 'MULTIPLY':
                col.prop(strip, "speed_factor", text=" ")
            elif strip.speed_control == 'LENGTH':
                col.prop(strip, "speed_length", text=" ")
            elif strip.speed_control == 'FRAME_NUMBER':
                col.prop(strip, "speed_frame_number", text=" ")

            row = layout.row(align=True)
            row.enabled = strip.speed_control != 'STRETCH'
            row = layout.row(align=True, heading="Interpolation")
            row.prop(strip, "use_frame_interpolate", text="")

        elif strip_type == 'MULTICAM':
            col = layout.column(align=True)
            strip_channel = strip.channel

            col.prop(strip, "multicam_source", text="Source Channel")

            # The multicam strip needs at least 2 strips to be useful
            if strip_channel > 2:
                BT_ROW = 4
                col.label(text="Cut To")
                row = col.row()

                for i in range(1, strip_channel):
                    if (i % BT_ROW) == 1:
                        row = col.row(align=True)

                    # Workaround - .enabled has to have a separate UI block to work
                    if i == strip.multicam_source:
                        sub = row.row(align=True)
                        sub.enabled = False
                        sub.operator("sequencer.split_multicam", text="{:d}".format(i), translate=False).camera = i
                    else:
                        sub_1 = row.row(align=True)
                        sub_1.enabled = True
                        sub_1.operator("sequencer.split_multicam", text="{:d}".format(i), translate=False).camera = i

                if strip.channel > BT_ROW and (strip_channel - 1) % BT_ROW:
                    for i in range(strip.channel, strip_channel + ((BT_ROW + 1 - strip_channel) % BT_ROW)):
                        row.label(text="")
            else:
                col.separator()
                col.label(text="Two or more channels are needed below this strip", icon='INFO')

        elif strip_type == 'TEXT':
            layout = self.layout
            col = layout.column()
            col.scale_x = 1.3
            col.scale_y = 1.3
            col.use_property_split = False
            col.prop(strip, "text", text="")
            col.use_property_split = True
            layout.prop(strip, "wrap_width", text="Wrap Width")

        col = layout.column(align=True)
        if strip_type in {'CROSS', 'GAMMA_CROSS', 'WIPE', 'ALPHA_OVER', 'ALPHA_UNDER'}:
            col.prop(strip, "use_default_fade", text="Default Fade")
            if not strip.use_default_fade:
                col.prop(strip, "effect_fader", text="Effect Fader")
        elif strip_type == 'GAUSSIAN_BLUR':
            col = layout.column(align=True)
            col.prop(strip, "size_x", text="Size X")
            col.prop(strip, "size_y", text="Y")
        elif strip_type == 'COLORMIX':
            layout.prop(strip, "blend_effect", text="Blend Mode")
            row = layout.row(align=True)
            row.prop(strip, "factor", slider=True)


class STRIP_PT_effect_text_layout(StripButtonsPanel, Panel):
    bl_label = "Layout"
    bl_parent_id = "STRIP_PT_effect"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        return strip.type == 'TEXT'

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True
        col = layout.column()
        col.prop(strip, "location", text="Location")
        col.prop(strip, "alignment_x", text="Alignment X")
        col.prop(strip, "anchor_x", text="Anchor X")
        col.prop(strip, "anchor_y", text="Y")


class STRIP_PT_effect_text_style(StripButtonsPanel, Panel):
    bl_label = "Style"
    bl_parent_id = "STRIP_PT_effect"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        return strip.type == 'TEXT'

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True
        col = layout.column()

        row = col.row(align=True)
        row.use_property_decorate = False
        row.template_ID(strip, "font", open="font.open", unlink="font.unlink")
        row.prop(strip, "use_bold", text="", icon='BOLD')
        row.prop(strip, "use_italic", text="", icon='ITALIC')

        col.prop(strip, "font_size")
        col.prop(strip, "color")


class STRIP_PT_effect_text_outline(StripButtonsPanel, Panel):
    bl_label = "Outline"
    bl_options = {'DEFAULT_CLOSED'}

    bl_parent_id = "STRIP_PT_effect_text_style"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        return strip.type == 'TEXT'

    def draw_header(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.prop(strip, "use_outline", text="")

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True

        col = layout.column()
        col.prop(strip, "outline_color", text="Color")
        col.prop(strip, "outline_width", text="Width")
        col.active = strip.use_outline and (not strip.mute)


class STRIP_PT_effect_text_shadow(StripButtonsPanel, Panel):
    bl_label = "Shadow"
    bl_options = {'DEFAULT_CLOSED'}

    bl_parent_id = "STRIP_PT_effect_text_style"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        return strip.type == 'TEXT'

    def draw_header(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.prop(strip, "use_shadow", text="")

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True

        col = layout.column()
        col.prop(strip, "shadow_color", text="Color")
        col.prop(strip, "shadow_angle", text="Angle")
        col.prop(strip, "shadow_offset", text="Offset")
        col.prop(strip, "shadow_blur", text="Blur")
        col.active = strip.use_shadow and (not strip.mute)


class STRIP_PT_effect_text_box(StripButtonsPanel, Panel):
    bl_label = "Box"
    bl_translation_context = i18n_contexts.id_sequence
    bl_options = {'DEFAULT_CLOSED'}

    bl_parent_id = "STRIP_PT_effect_text_style"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        return strip.type == 'TEXT'

    def draw_header(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.prop(strip, "use_box", text="")

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True

        col = layout.column()
        col.prop(strip, "box_color", text="Color")
        col.prop(strip, "box_margin", text="Margin")
        col.prop(strip, "box_roundness", text="Roundness")
        col.active = strip.use_box and (not strip.mute)


class STRIP_PT_source(StripButtonsPanel, Panel):
    bl_label = "Source"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type in {'MOVIE', 'IMAGE', 'SOUND'}

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        scene = context.sequencer_scene
        strip = context.active_strip
        strip_type = strip.type

        layout.active = not strip.mute

        # Draw a filename if we have one.
        if strip_type == 'SOUND':
            sound = strip.sound
            layout.template_ID(strip, "sound", open="sound.open")
            if sound is not None:

                col = layout.column()
                col.prop(sound, "filepath", text="")

                col.alignment = 'RIGHT'
                sub = col.column(align=True)
                split = sub.split(factor=0.5, align=True)
                split.alignment = 'RIGHT'
                if sound.packed_file:
                    split.label(text="Unpack")
                    split.operator("sound.unpack", icon='PACKAGE', text="")
                else:
                    split.label(text="Pack")
                    split.operator("sound.pack", icon='UGLYPACKAGE', text="")

                layout.prop(sound, "use_memory_cache")

                col = layout.box()
                col = col.column(align=True)
                split = col.split(factor=0.5, align=False)
                split.alignment = 'RIGHT'
                split.label(text="Sample Rate")
                split.alignment = 'LEFT'
                if sound.samplerate <= 0:
                    split.label(text="Unknown")
                else:
                    split.label(text="{:d} Hz".format(sound.samplerate), translate=False)

                split = col.split(factor=0.5, align=False)
                split.alignment = 'RIGHT'
                split.label(text="Channels")
                split.alignment = 'LEFT'

                # FIXME(@campbellbarton): this is ugly, we may want to support a way of showing a label from an enum.
                channel_enum_items = sound.bl_rna.properties["channels"].enum_items
                split.label(text=channel_enum_items[channel_enum_items.find(sound.channels)].name)
                del channel_enum_items
        else:
            if strip_type == 'IMAGE':
                col = layout.column()
                col.prop(strip, "directory", text="")

                # Current element for the filename.
                elem = strip.strip_elem_from_frame(scene.frame_current)
                if elem:
                    col.prop(elem, "filename", text="")  # strip.elements[0] could be a fallback

                col.prop(strip.colorspace_settings, "name", text="Color Space")

                col.prop(strip, "alpha_mode", text="Alpha")
                sub = col.column(align=True)
                sub.operator("sequencer.change_path", text="Change Data/Files", icon='FILEBROWSER').filter_image = True
            else:  # elif strip_type == 'MOVIE':
                elem = strip.elements[0]

                col = layout.column()
                col.prop(strip, "filepath", text="")
                col.prop(strip.colorspace_settings, "name", text="Color Space")
                col.prop(strip, "stream_index")
                col.prop(strip, "use_deinterlace")

            if scene.render.use_multiview:
                layout.prop(strip, "use_multiview")

                col = layout.column()
                col.active = strip.use_multiview

                col.row().prop(strip, "views_format", expand=True)

                box = col.box()
                box.active = strip.views_format == 'STEREO_3D'
                box.template_image_stereo_3d(strip.stereo_3d_format)

            # Resolution.
            col = layout.box()
            col = col.column(align=True)
            split = col.split(factor=0.5, align=False)
            split.alignment = 'RIGHT'
            split.label(text="Resolution")
            size = (elem.orig_width, elem.orig_height) if elem else (0, 0)
            if size[0] and size[1]:
                split.alignment = 'LEFT'
                split.label(text="{:d}x{:d}".format(*size), translate=False)
            else:
                split.label(text="None")
            # FPS
            if elem.orig_fps:
                split = col.split(factor=0.5, align=False)
                split.alignment = 'RIGHT'
                split.label(text="FPS")
                split.alignment = 'LEFT'
                split.label(text="{:.2f}".format(elem.orig_fps), translate=False)


class STRIP_PT_movie_clip(StripButtonsPanel, Panel):
    bl_label = "Movie Clip"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type == 'MOVIECLIP'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False

        strip = context.active_strip

        layout.active = not strip.mute
        layout.template_ID(strip, "clip")

        if strip.type == 'MOVIECLIP':
            col = layout.column(heading="Use")
            col.prop(strip, "stabilize2d", text="2D Stabilized Clip")
            col.prop(strip, "undistort", text="Undistorted Clip")

        clip = strip.clip
        if clip:
            sta = clip.frame_start
            end = clip.frame_start + clip.frame_duration
            layout.label(
                text=rpt_("Original frame range: {:d}-{:d} ({:d})").format(sta, end, end - sta + 1),
                translate=False,
            )


class STRIP_PT_scene(StripButtonsPanel, Panel):
    bl_label = "Scene"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return (strip.type == 'SCENE')

    def draw(self, context):
        strip = context.active_strip
        scene = strip.scene

        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = not strip.mute

        layout.template_ID(strip, "scene", text="Scene", new="scene.new_sequencer")
        layout.prop(strip, "scene_input", text="Input")

        if strip.scene_input == 'CAMERA':
            layout.template_ID(strip, "scene_camera", text="Camera")

        if strip.scene_input == 'CAMERA':
            layout = layout.column(heading="Show")
            layout.prop(strip, "use_annotations", text="Annotations")
            if scene:
                # Warning, this is not a good convention to follow.
                # Expose here because setting the alpha from the "Render" menu is very inconvenient.
                layout.prop(scene.render, "film_transparent")


class STRIP_PT_scene_sound(StripButtonsPanel, Panel):
    bl_label = "Sound"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return (strip.type == 'SCENE')

    def draw(self, context):
        strip = context.active_strip

        layout = self.layout
        layout.use_property_split = True
        layout.use_property_decorate = False
        layout.active = not strip.mute

        col = layout.column()

        col.use_property_decorate = True
        split = col.split(factor=0.4)
        split.alignment = 'RIGHT'
        split.label(text="Strip Volume", text_ctxt=i18n_contexts.id_sound)
        split.prop(strip, "volume", text="")
        col.use_property_decorate = False


class STRIP_PT_mask(StripButtonsPanel, Panel):
    bl_label = "Mask"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return (strip.type == 'MASK')

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        strip = context.active_strip

        layout.active = not strip.mute

        layout.template_ID(strip, "mask")

        mask = strip.mask

        if mask:
            sta = mask.frame_start
            end = mask.frame_end
            layout.label(
                text=rpt_("Original frame range: {:d}-{:d} ({:d})").format(sta, end, end - sta + 1),
                translate=False,
            )


class STRIP_PT_time(StripButtonsPanel, Panel):
    bl_label = "Time"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type

    def draw_header_preset(self, context):
        layout = self.layout
        layout.alignment = 'RIGHT'
        strip = context.active_strip

        layout.prop(strip, "lock", text="", icon_only=True, emboss=False)

    def draw(self, context):
        from bpy.utils import smpte_from_frame

        layout = self.layout
        layout.use_property_split = False
        layout.use_property_decorate = False

        scene = context.sequencer_scene
        frame_current = scene.frame_current
        strip = context.active_strip

        is_effect = isinstance(strip, bpy.types.EffectStrip)

        # Get once.
        frame_start = strip.frame_start
        frame_final_start = strip.frame_final_start
        frame_final_end = strip.frame_final_end
        frame_final_duration = strip.frame_final_duration
        frame_offset_start = strip.frame_offset_start
        frame_offset_end = strip.frame_offset_end

        length_list = (
            str(round(frame_start, 0)),
            str(round(frame_final_end, 0)),
            str(round(frame_final_duration, 0)),
            str(round(frame_offset_start, 0)),
            str(round(frame_offset_end, 0)),
        )

        if not is_effect:
            length_list = length_list + (
                str(round(strip.animation_offset_start, 0)),
                str(round(strip.animation_offset_end, 0)),
            )

        max_length = max(len(x) for x in length_list)
        max_factor = (1.9 - max_length) / 30
        factor = 0.45

        layout.enabled = not strip.lock
        layout.active = not strip.mute

        sub = layout.row(align=True)
        split = sub.split(factor=factor + max_factor)
        split.alignment = 'RIGHT'
        split.label(text="")
        split.prop(strip, "show_retiming_keys")

        sub = layout.row(align=True)
        split = sub.split(factor=factor + max_factor)
        split.alignment = 'RIGHT'
        split.label(text="Channel")
        split.prop(strip, "channel", text="")

        sub = layout.column(align=True)
        split = sub.split(factor=factor + max_factor, align=True)
        split.alignment = 'RIGHT'
        split.label(text="Start")
        split.prop(strip, "frame_start", text=smpte_from_frame(frame_start))

        split = sub.split(factor=factor + max_factor, align=True)
        split.alignment = 'RIGHT'
        split.label(text="Duration")
        split.prop(strip, "frame_final_duration", text=smpte_from_frame(frame_final_duration))

        # Use label, editing this value from the UI allows negative values,
        # users can adjust duration.
        split = sub.split(factor=factor + max_factor, align=True)
        split.alignment = 'RIGHT'
        split.label(text="End")
        split = split.split(factor=factor + 0.3 + max_factor, align=True)
        split.label(text="{:>14s}".format(smpte_from_frame(frame_final_end)), translate=False)
        split.alignment = 'RIGHT'
        split.label(text=str(frame_final_end) + " ")

        if not is_effect:

            layout.alignment = 'RIGHT'
            sub = layout.column(align=True)

            split = sub.split(factor=factor + max_factor, align=True)
            split.alignment = 'RIGHT'
            split.label(text="Strip Offset Start")
            split.prop(strip, "frame_offset_start", text=smpte_from_frame(frame_offset_start))

            split = sub.split(factor=factor + max_factor, align=True)
            split.alignment = 'RIGHT'
            split.label(text="End")
            split.prop(strip, "frame_offset_end", text=smpte_from_frame(frame_offset_end))

            layout.alignment = 'RIGHT'
            sub = layout.column(align=True)

            split = sub.split(factor=factor + max_factor, align=True)
            split.alignment = 'RIGHT'
            split.label(text="Hold Offset Start")
            split.prop(strip, "animation_offset_start", text=smpte_from_frame(strip.animation_offset_start))

            split = sub.split(factor=factor + max_factor, align=True)
            split.alignment = 'RIGHT'
            split.label(text="End")
            split.prop(strip, "animation_offset_end", text=smpte_from_frame(strip.animation_offset_end))

            if strip.type == 'SOUND':
                sub2 = layout.column(align=True)
                split = sub2.split(factor=factor + max_factor, align=True)
                split.alignment = 'RIGHT'
                split.label(text="Sound Offset", text_ctxt=i18n_contexts.id_sound)
                split.prop(strip, "sound_offset", text="")

        col = layout.column(align=True)
        col = col.box()
        col.active = (
            (frame_current >= frame_final_start) and
            (frame_current <= frame_final_start + frame_final_duration)
        )

        split = col.split(factor=factor + max_factor, align=True)
        split.alignment = 'RIGHT'
        split.label(text="Current Frame")
        split = split.split(factor=factor + 0.3 + max_factor, align=True)
        frame_display = frame_current - frame_final_start
        split.label(text="{:>14s}".format(smpte_from_frame(frame_display)), translate=False)
        split.alignment = 'RIGHT'
        split.label(text=str(frame_display) + " ")

        if strip.type == 'SCENE':
            scene = strip.scene

            if scene:
                sta = scene.frame_start
                end = scene.frame_end
                split = col.split(factor=factor + max_factor)
                split.alignment = 'RIGHT'
                split.label(text="Original Frame Range")
                split.alignment = 'LEFT'
                split.label(text="{:d}-{:d} ({:d})".format(sta, end, end - sta + 1), translate=False)


class STRIP_PT_adjust_sound(StripButtonsPanel, Panel):
    bl_label = "Sound"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type == 'SOUND'

    def draw(self, context):
        layout = self.layout

        strip = context.active_strip
        sound = strip.sound

        layout.active = not strip.mute

        if sound is not None:
            layout.use_property_split = True
            col = layout.column()

            split = col.split(factor=0.4)
            split.alignment = 'RIGHT'
            split.label(text="Volume", text_ctxt=i18n_contexts.id_sound)
            split.prop(strip, "volume", text="")

            layout.use_property_split = False
            col = layout.column()

            split = col.split(factor=0.4)
            split.label(text="")
            split.prop(sound, "use_mono")

            layout.use_property_split = True
            col = layout.column()

            audio_channels = context.sequencer_scene.render.ffmpeg.audio_channels
            pan_enabled = sound.use_mono and audio_channels != 'MONO'
            pan_text = "{:.2f}°".format(strip.pan * 90.0)

            split = col.split(factor=0.4)
            split.alignment = 'RIGHT'
            split.label(text="Pan", text_ctxt=i18n_contexts.id_sound)
            split.prop(strip, "pan", text="")
            split.enabled = pan_enabled

            if audio_channels not in {'MONO', 'STEREO'}:
                split = col.split(factor=0.4)
                split.alignment = 'RIGHT'
                split.label(text="Pan Angle")
                split.enabled = pan_enabled
                subsplit = split.row()
                subsplit.alignment = 'CENTER'
                subsplit.label(text=pan_text)
                subsplit.label(text=" ")  # Compensate for no decorate.
                subsplit.enabled = pan_enabled

            layout.use_property_split = False
            col = layout.column()

            split = col.split(factor=0.4)
            split.label(text="")
            split.prop(strip, "pitch_correction")

            split = col.split(factor=0.4)
            split.label(text="")
            split.prop(strip, "show_waveform")


class STRIP_PT_adjust_comp(StripButtonsPanel, Panel):
    bl_label = "Compositing"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type != 'SOUND'

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        strip = context.active_strip

        layout.active = not strip.mute

        col = layout.column()
        col.prop(strip, "blend_type", text="Blend")
        col.prop(strip, "blend_alpha", text="Opacity", slider=True)


class STRIP_PT_adjust_transform(StripButtonsPanel, Panel):
    bl_label = "Transform"

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type != 'SOUND'

    def draw(self, context):
        strip = context.active_strip
        layout = self.layout
        layout.use_property_split = True
        layout.active = not strip.mute

        col = layout.column(align=True)
        col.prop(strip.transform, "filter", text="Filter")

        col = layout.column(align=True)
        col.prop(strip.transform, "offset_x", text="Position X")
        col.prop(strip.transform, "offset_y", text="Y")

        col = layout.column(align=True)
        col.prop(strip.transform, "scale_x", text="Scale X")
        col.prop(strip.transform, "scale_y", text="Y")

        col = layout.column(align=True)
        col.prop(strip.transform, "rotation", text="Rotation")

        col = layout.column(align=True)
        col.prop(strip.transform, "origin")

        col = layout.column(heading="Mirror", align=True, heading_ctxt=i18n_contexts.id_image)
        col.prop(strip, "use_flip_x", text="X", toggle=True)
        col.prop(strip, "use_flip_y", text="Y", toggle=True)


class STRIP_PT_adjust_video(StripButtonsPanel, Panel):
    bl_label = "Video"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type in {
            'MOVIE', 'IMAGE', 'SCENE', 'MOVIECLIP', 'MASK',
            'META', 'ADD', 'SUBTRACT', 'ALPHA_OVER',
            'ALPHA_UNDER', 'CROSS', 'GAMMA_CROSS', 'MULTIPLY',
            'WIPE', 'GLOW', 'COLOR', 'MULTICAM', 'SPEED', 'ADJUSTMENT', 'COLORMIX',
        }

    def draw(self, context):
        layout = self.layout

        layout.use_property_split = True

        col = layout.column()

        strip = context.active_strip

        layout.active = not strip.mute

        col.prop(strip, "strobe")
        col.prop(strip, "use_reverse_frames")


class STRIP_PT_adjust_color(StripButtonsPanel, Panel):
    bl_label = "Color"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        strip = context.active_strip
        if not strip:
            return False

        return strip.type in {
            'MOVIE', 'IMAGE', 'SCENE', 'MOVIECLIP', 'MASK',
            'META', 'ADD', 'SUBTRACT', 'ALPHA_OVER',
            'ALPHA_UNDER', 'CROSS', 'GAMMA_CROSS', 'MULTIPLY',
            'WIPE', 'GLOW', 'COLOR', 'MULTICAM', 'SPEED', 'ADJUSTMENT', 'COLORMIX',
        }

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        strip = context.active_strip

        layout.active = not strip.mute

        col = layout.column()
        col.prop(strip, "color_saturation", text="Saturation")
        col.prop(strip, "color_multiply", text="Multiply")
        col.prop(strip, "multiply_alpha")
        col.prop(strip, "use_float", text="Convert to Float")


class STRIP_PT_custom_props(StripButtonsPanel, PropertyPanel, Panel):
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_WORKBENCH',
    }
    _context_path = "active_strip"
    _property_type = (bpy.types.Strip,)


classes = (
    STRIP_PT_color_tag_picker,

    STRIP_PT_strip,
    STRIP_PT_effect,
    STRIP_PT_scene,
    STRIP_PT_scene_sound,
    STRIP_PT_mask,
    STRIP_PT_effect_text_style,
    STRIP_PT_effect_text_outline,
    STRIP_PT_effect_text_shadow,
    STRIP_PT_effect_text_box,
    STRIP_PT_effect_text_layout,
    STRIP_PT_movie_clip,

    STRIP_PT_adjust_comp,
    STRIP_PT_adjust_transform,
    STRIP_PT_adjust_crop,
    STRIP_PT_adjust_video,
    STRIP_PT_adjust_color,
    STRIP_PT_adjust_sound,

    STRIP_PT_time,
    STRIP_PT_source,

    STRIP_PT_custom_props,
)

if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



# SPDX-FileCopyrightText: 2018-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import bpy
from bpy.props import (
    BoolProperty,
    EnumProperty,
)
from bpy.app.translations import contexts as i18n_contexts

DIRNAME, FILENAME = os.path.split(__file__)
IDNAME = os.path.splitext(FILENAME)[0]


def update_fn(_self, _context):
    load()


class Prefs(bpy.types.KeyConfigPreferences):
    bl_idname = IDNAME

    select_mouse: EnumProperty(
        name="Select Mouse",
        items=(
            ('LEFT', "Left",
             "Use left mouse button for selection. "
             "The standard behavior that works well for mouse, trackpad and tablet devices"),
            ('RIGHT', "Right",
             "Use right mouse button for selection, and left mouse button for actions. "
             "This works well primarily for keyboard and mouse devices"),
        ),
        description=(
            "Mouse button used for selection"
        ),
        update=update_fn,
    )
    spacebar_action: EnumProperty(
        name="Spacebar Action",
        items=(
            ('PLAY', "Play",
             "Toggle animation playback "
             "('Shift-Space' for Tools or brush asset popup)",
             1),
            ('TOOL', "Tools",
             "Open the popup tool-bar or brush asset popup\n"
             "When holding 'Space' for the tool-bar popup:\n"
             "\u2022 Pressing the a tool shortcut switches to it immediately.\n"
             "\u2022 Dragging the cursor over a tool and releasing activates it (like a pie menu).\n"
             "For Play use 'Shift-Space'",
             0),
            ('SEARCH', "Search",
             "Open the operator search popup",
             2),
        ),
        description=(
            "Action when 'Space' is pressed"
        ),
        default='PLAY',
        update=update_fn,
    )
    tool_key_mode: EnumProperty(
        name="Tool Keys",
        description=(
            "The method of keys to activate tools such as move, rotate & scale (G, R, S)"
        ),
        items=(
            ('IMMEDIATE', "Immediate",
             "Activate actions immediately"),
            ('TOOL', "Active Tool",
             "Activate the tool for editors that support tools"),
        ),
        default='IMMEDIATE',
        update=update_fn,
    )

    rmb_action: EnumProperty(
        name="Right Mouse Select Action",
        items=(
            ('TWEAK', "Select & Tweak",
             "Right mouse always tweaks"),
            ('FALLBACK_TOOL', "Selection Tool",
             "Right mouse uses the selection tool"),
        ),
        description=(
            "Default action for the right mouse button"
        ),
        update=update_fn,
    )

    use_region_toggle_pie: BoolProperty(
        name="Region Toggle Pie",
        description=(
            "N-key opens a pie menu to toggle regions"
        ),
        default=False,
        update=update_fn,
    )

    use_alt_click_leader: BoolProperty(
        name="Alt Click Tool Prompt",
        description=(
            "Tapping Alt (without pressing any other keys) shows a prompt in the status-bar, "
            "prompting a second keystroke to activate the tool"
        ),
        default=False,
        update=update_fn,
    )
    # NOTE: expose `use_alt_tool` and `use_alt_cursor` as two options in the UI
    # as the tool-tips and titles are different enough depending on RMB/LMB select.
    use_alt_tool: BoolProperty(
        name="Alt Tool Access",
        description=(
            "Hold Alt to use the active tool when the gizmo would normally be required\n"
            "Incompatible with the input preference \"Emulate 3 Button Mouse\" when the \"Alt\" key is used"
        ),
        default=False,
        update=update_fn,
    )
    use_alt_cursor: BoolProperty(
        name="Alt Cursor Access",
        description=(
            "Hold Alt-LMB to place the Cursor (instead of LMB), allows tools to activate on press instead of drag.\n"
            "Incompatible with the input preference \"Emulate 3 Button Mouse\" when the \"Alt\" key is used"
        ),
        default=False,
        update=update_fn,
    )
    # end note.

    use_select_all_toggle: BoolProperty(
        name="Select All Toggles",
        description=(
            "Causes select-all ('A' key) to de-select in the case a selection exists"
        ),
        default=False,
        update=update_fn,
    )

    gizmo_action: EnumProperty(
        name="Activate Gizmo",
        translation_context=i18n_contexts.editor_view3d,
        items=(
            ('PRESS', "Press", "Press causes immediate activation, preventing click being passed to the tool"),
            ('DRAG', "Drag", "Drag allows click events to pass through to the tool, adding a small delay"),
        ),
        description="Activation event for gizmos that support drag motion",
        default='DRAG',
        update=update_fn,
    )

    # 3D View
    use_v3d_tab_menu: BoolProperty(
        name="Tab for Pie Menu",
        description=(
            "Causes tab to open pie menu (swaps 'Tab' / 'Ctrl-Tab')"
        ),
        default=False,
        update=update_fn,
    )
    use_v3d_shade_ex_pie: BoolProperty(
        name="Extra Shading Pie Menu Items",
        description=(
            "Show additional options in the shading menu ('Z')"
        ),
        default=False,
        update=update_fn,
    )
    v3d_tilde_action: EnumProperty(
        name="Tilde Action",
        items=(
            ('VIEW', "Navigate",
             "View operations (useful for keyboards without a numpad)",
             0),
            ('GIZMO', "Gizmos",
             "Control transform gizmos",
             1),
        ),
        description=(
            "Action when 'Tilde' is pressed"
        ),
        default='VIEW',
        update=update_fn,
    )

    v3d_mmb_action: EnumProperty(
        name="MMB Action",
        items=(
            ('ORBIT', "Orbit",
             "",
             0),
            ('PAN', "Pan",
             "",
             1),
        ),
        description=(
            "The action when Middle-Mouse dragging in the viewport. "
            "Shift-Middle-Mouse is used for the other action. "
            "This applies to trackpad as well"
        ),
        update=update_fn,
    )

    v3d_alt_mmb_drag_action: EnumProperty(
        name="Alt-MMB Drag Action",
        items=(
            ('RELATIVE', "Relative",
             "Set the view axis where each mouse direction maps to an axis relative to the current orientation",
             0),
            ('ABSOLUTE', "Absolute",
             "Set the view axis where each mouse direction always maps to the same axis",
             1),
        ),
        description=(
            "Action when Alt-MMB dragging in the 3D viewport"
        ),
        update=update_fn,
    )

    # Developer note, this is an experimental option.
    use_pie_click_drag: BoolProperty(
        name="Pie Menu on Drag",
        description=(
            "Activate some pie menus on drag,\n"
            "allowing the tapping the same key to have a secondary action.\n"
            "\n"
            "\u2022 Tapping Tab in the 3D view toggles edit-mode, drag for mode menu.\n"
            "\u2022 Tapping Z in the 3D view toggles wireframe, drag for draw modes.\n"
            "\u2022 Tapping Tilde in the 3D view for first person navigation, drag for view axes"
        ),
        default=False,
        update=update_fn,
    )

    use_file_single_click: BoolProperty(
        name="Open Folders on Single Click",
        description=(
            "Navigate into folders by clicking on them once instead of twice"
        ),
        default=False,
        update=update_fn,
    )

    use_alt_navigation: BoolProperty(
        name="Transform Navigation with Alt",
        description=(
            "During transformations, use Alt to navigate in the 3D View. "
            "Note that if disabled, hotkeys for Proportional Editing, "
            "Automatic Constraints, and Auto IK Chain Length will require holding Alt"
        ),
        default=True,
        update=update_fn,
    )

    def draw(self, layout):
        from bpy import context

        layout.use_property_split = True
        layout.use_property_decorate = False

        prefs = context.preferences

        show_developer_ui = prefs.view.show_developer_ui
        is_select_left = (self.select_mouse == 'LEFT')
        use_mouse_emulate_3_button = (
            prefs.inputs.use_mouse_emulate_3_button and
            prefs.inputs.mouse_emulate_3_button_modifier == 'ALT'
        )

        # General settings.
        col = layout.column()
        col.row().prop(self, "select_mouse", text="Select with Mouse Button", expand=True)
        col.row().prop(self, "spacebar_action", text="Spacebar Action", expand=True)

        if is_select_left:
            col.row().prop(self, "gizmo_action", text="Activate Gizmo Event", expand=True)
        else:
            col.row().prop(self, "rmb_action", text="Right Mouse Select Action", expand=True)

        col.row().prop(self, "tool_key_mode", expand=True)

        # Check-box sub-layout.
        col = layout.column()
        sub = col.column(align=True)
        row = sub.row()
        row.prop(self, "use_alt_click_leader")

        rowsub = row.row()
        if is_select_left:
            rowsub.prop(self, "use_alt_tool")
        else:
            rowsub.prop(self, "use_alt_cursor")
        rowsub.active = not use_mouse_emulate_3_button

        row = sub.row()
        row.prop(self, "use_select_all_toggle")

        row = sub.row()
        row.prop(self, "use_region_toggle_pie")

        # 3DView settings.
        col = layout.column()
        col.label(text="3D View")
        col.row().prop(self, "v3d_tilde_action", text="Grave Accent / Tilde Action", expand=True)
        col.row().prop(self, "v3d_mmb_action", text="Middle Mouse Action", expand=True)
        col.row().prop(self, "v3d_alt_mmb_drag_action", text="Alt Middle Mouse Drag Action", expand=True)

        # Check-boxes sub-layout.
        col = layout.column()
        sub = col.column(align=True)
        sub.prop(self, "use_v3d_tab_menu")
        sub.prop(self, "use_pie_click_drag")
        sub.prop(self, "use_v3d_shade_ex_pie")
        sub.prop(self, "use_alt_navigation")

        # File Browser settings.
        col = layout.column()
        col.label(text="File Browser")
        col.row().prop(self, "use_file_single_click")


blender_default = bpy.utils.execfile(os.path.join(DIRNAME, "keymap_data", "blender_default.py"))


def load():
    from sys import platform
    from bpy import context
    from bl_keymap_utils.io import keyconfig_init_from_data

    prefs = context.preferences
    kc = context.window_manager.keyconfigs.new(IDNAME)
    kc_prefs = kc.preferences

    show_developer_ui = prefs.view.show_developer_ui
    is_select_left = (kc_prefs.select_mouse == 'LEFT')
    use_mouse_emulate_3_button = (
        prefs.inputs.use_mouse_emulate_3_button and
        prefs.inputs.mouse_emulate_3_button_modifier == 'ALT'
    )

    keyconfig_data = blender_default.generate_keymaps(
        blender_default.Params(
            select_mouse=kc_prefs.select_mouse,
            use_mouse_emulate_3_button=use_mouse_emulate_3_button,
            spacebar_action=kc_prefs.spacebar_action,
            use_key_activate_tools=(kc_prefs.tool_key_mode == 'TOOL'),
            use_region_toggle_pie=kc_prefs.use_region_toggle_pie,
            v3d_tilde_action=kc_prefs.v3d_tilde_action,
            use_v3d_mmb_pan=(kc_prefs.v3d_mmb_action == 'PAN'),
            v3d_alt_mmb_drag_action=kc_prefs.v3d_alt_mmb_drag_action,
            use_select_all_toggle=kc_prefs.use_select_all_toggle,
            use_v3d_tab_menu=kc_prefs.use_v3d_tab_menu,
            use_v3d_shade_ex_pie=kc_prefs.use_v3d_shade_ex_pie,
            use_gizmo_drag=(is_select_left and kc_prefs.gizmo_action == 'DRAG'),
            use_fallback_tool=True,
            use_fallback_tool_select_handled=(
                # LMB doesn't need additional selection fallback key-map items.
                False if is_select_left else
                # RMB is select and RMB must trigger the fallback tool.
                # Otherwise LMB activates the fallback tool and RMB always tweak-selects.
                (kc_prefs.rmb_action != 'FALLBACK_TOOL')
            ),
            use_alt_tool_or_cursor=(
                (not use_mouse_emulate_3_button) and
                (kc_prefs.use_alt_tool if is_select_left else kc_prefs.use_alt_cursor)
            ),
            use_alt_click_leader=kc_prefs.use_alt_click_leader,
            use_pie_click_drag=kc_prefs.use_pie_click_drag,
            use_file_single_click=kc_prefs.use_file_single_click,
            use_alt_navigation=kc_prefs.use_alt_navigation,
        ),
    )

    if platform == "darwin":
        from bl_keymap_utils.platform_helpers import keyconfig_data_oskey_from_ctrl_for_macos
        keyconfig_data = keyconfig_data_oskey_from_ctrl_for_macos(keyconfig_data)

    keyconfig_init_from_data(kc, keyconfig_data)


if __name__ == "__main__":
    bpy.utils.register_class(Prefs)
    load()



# SPDX-FileCopyrightText: 2019-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from itertools import count

from ...utils.naming import strip_org, make_derived_name
from ...utils.bones import put_bone, set_bone_widget_transform
from ...utils.widgets_basic import create_circle_widget
from ...utils.misc import map_list

from ...base_rig import stage

from ..widgets import create_ball_socket_widget

from .spine_rigs import BaseHeadTailRig


class Rig(BaseHeadTailRig):
    copy_rotation_axes: tuple[bool, bool, bool]

    def initialize(self):
        super().initialize()

        self.copy_rotation_axes = self.params.copy_rotation_axes

    def parent_bones(self):
        super().parent_bones()

        if self.connected_tweak and self.use_connect_reverse:
            self.rig_parent_bone = self.connected_tweak

    ####################################################
    # BONES

    class CtrlBones(BaseHeadTailRig.CtrlBones):
        master: str                    # Master control.

    class MchBones(BaseHeadTailRig.MchBones):
        rot_tail: str                  # Tail follow system.

    bones: BaseHeadTailRig.ToplevelBones[
        list[str],
        'Rig.CtrlBones',
        'Rig.MchBones',
        list[str]
    ]

    ####################################################
    # Master control

    @stage.generate_bones
    def make_master_control(self):
        org = self.bones.org[0]
        self.bones.ctrl.master = self.copy_bone(org, make_derived_name(org, 'ctrl', '_master'))
        self.default_prop_bone = self.bones.ctrl.master

    @stage.parent_bones
    def parent_master_control(self):
        self.set_bone_parent(self.bones.ctrl.master, self.rig_parent_bone)

    @stage.configure_bones
    def configure_master_control(self):
        bone = self.get_bone(self.bones.ctrl.master)
        bone.lock_location = True, True, True

    @stage.generate_widgets
    def make_master_control_widget(self):
        bone = self.bones.ctrl.master
        set_bone_widget_transform(self.obj, bone, self.bones.ctrl.tweak[-1])
        create_ball_socket_widget(self.obj, bone, size=0.7)

    ####################################################
    # Control bones

    @stage.parent_bones
    def parent_control_chain(self):
        self.set_bone_parent(self.bones.ctrl.fk[0], self.bones.mch.rot_tail)
        self.parent_bone_chain(self.bones.ctrl.fk, use_connect=False)

    @stage.rig_bones
    def rig_control_chain(self):
        ctrls = self.bones.ctrl.fk
        for args in zip(count(0), ctrls, [self.bones.ctrl.master] + ctrls):
            self.rig_control_bone(*args)

    def rig_control_bone(self, _i: int, ctrl: str, prev_ctrl: str):
        self.make_constraint(
            ctrl, 'COPY_ROTATION', prev_ctrl,
            use_xyz=self.copy_rotation_axes,
            space='LOCAL', mix_mode='BEFORE',
        )

    # Widgets
    def make_control_widget(self, i: int, ctrl: str):
        create_circle_widget(self.obj, ctrl, radius=0.5, head_tail=0.75)

    ####################################################
    # MCH bones associated with main controls

    @stage.generate_bones
    def make_mch_control_bones(self):
        self.bones.mch.rot_tail = self.make_mch_follow_bone(self.bones.org[0], 'tail', 0.0)

    @stage.parent_bones
    def parent_mch_control_bones(self):
        self.set_bone_parent(self.bones.mch.rot_tail, self.rig_parent_bone)

    ####################################################
    # Tweak bones

    @stage.generate_bones
    def make_tweak_chain(self):
        orgs = self.bones.org
        self.bones.ctrl.tweak = map_list(self.make_tweak_bone, count(0), orgs[0:1] + orgs)

    def make_tweak_bone(self, i: int, org: str):
        if i == 0:
            if self.check_connect_tweak(org):
                return self.connected_tweak

            else:
                name = self.copy_bone(org, 'tweak_base_' + strip_org(org), parent=False, scale=0.5)

        else:
            name = self.copy_bone(org, 'tweak_' + strip_org(org), parent=False, scale=0.5)
            put_bone(self.obj, name, self.get_bone(org).tail)

        return name

    ####################################################
    # Deform chain

    @stage.configure_bones
    def configure_deform_chain(self):
        if self.use_connect_chain and self.use_connect_reverse:
            self.get_bone(self.bones.deform[-1]).bone.bbone_easein = 0.0
            self.get_bone(self.rigify_parent.bones.deform[0]).bone.bbone_easein = 1.0
        else:
            self.get_bone(self.bones.deform[-1]).bone.bbone_easeout = 0.0

    ####################################################
    # SETTINGS

    @classmethod
    def add_parameters(cls, params):
        """ Add the parameters of this rig type to the
            RigifyParameters PropertyGroup
        """

        super().add_parameters(params)

        params.copy_rotation_axes = bpy.props.BoolVectorProperty(
            size=3,
            description="Automation axes",
            default=tuple([i == 0 for i in range(0, 3)])
        )

    @classmethod
    def parameters_ui(cls, layout, params):
        """ Create the ui for the rig parameters.
        """

        row = layout.row(align=True)
        for i, axis in enumerate(['x', 'y', 'z']):
            row.prop(params, "copy_rotation_axes", index=i, toggle=True, text=axis)

        super().parameters_ui(layout, params)


def create_sample(obj, *, parent=None):
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('tail')
    bone.head[:] = 0.0000, 0.0552, 1.0099
    bone.tail[:] = -0.0000, 0.0582, 0.8669
    bone.roll = 0.0000
    bone.use_connect = False
    if parent:
        bone.parent = arm.edit_bones[parent]
    bones['tail'] = bone.name
    bone = arm.edit_bones.new('tail.001')
    bone.head[:] = -0.0000, 0.0582, 0.8669
    bone.tail[:] = -0.0000, 0.0365, 0.7674
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['tail']]
    bones['tail.001'] = bone.name
    bone = arm.edit_bones.new('tail.002')
    bone.head[:] = -0.0000, 0.0365, 0.7674
    bone.tail[:] = -0.0000, 0.0010, 0.6984
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['tail.001']]
    bones['tail.002'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['tail']]
    pbone.rigify_type = 'spines.basic_tail'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.connect_chain = bool(parent)
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['tail.001']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['tail.002']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone
        if bcoll := arm.collections.active:
            bcoll.assign(bone)
